# pylint: skip-file

from collections.abc import Mapping, Sequence
from typing import List

__all__: List[str] = []


class DataObject:
    """
    An object for holding data as attributes.

    ``DataObject`` can be instantiated like ``dict``::

        >>> d = DataObject({'one': 1, 'two': 2}, three=3)
        >>> d.__dict__
        {'one': 1, 'two': 2, 'three': 3}

    Attributes are accessible via the regular dot notation (``d.x``) as well as
    the subscription syntax (``d['x']``)::

        >>> d.one == d['one'] == 1
        True

    To convert a ``DataObject`` into a dictionary, use ``d._to_dict()``.

    ``DataObject`` implements the following collection-like operations:

        * iteration through attributes as name-value pairs
        * ``'x' in d`` for membership tests
        * ``len(d)`` to get the number of attributes

    Additionally, the following methods are equivalent to their ``dict` counterparts:
    ``_clear``, ``_get``, ``_keys``, ``_items``, ``_pop``, ``_setdefault``, ``_update``.

    An advantage of ``DataObject`` over ``dict` subclasses is that every method name
    in ``DataObject`` begins with an underscore, so attributes like ``"update"`` or
    ``"values"`` are valid.
    """

    def __init__(self, *args, **kwargs):
        source = args[0] if args else {}
        self._update(source, **kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.__dict__!r})"

    def _copy(self):
        return self.__class__(self)

    __copy__ = _copy

    def __eq__(self, other):
        return isinstance(other, DataObject) and self.__dict__ == other.__dict__

    def __iter__(self):
        return iter(self.__dict__.items())

    def _update(self, source=None, **kwargs):
        if isinstance(source, DataObject):
            source = source.__dict__
        self.__dict__.update(source, **kwargs)

    def _setdefaults(self, source):
        if isinstance(source, dict):
            source = source.items()
        for name, value in source:
            self._setdefault(name, value)
        return self

    def _to_dict(self):
        d = dict(self.__dict__)
        for k, v in d.items():
            if isinstance(v, DataObject):
                d[k] = v._to_dict()
        return d

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __delitem__(self, key):
        del self.__dict__[key]

    def __len__(self):
        return len(self.__dict__)

    def __contains__(self, key):
        return key in self.__dict__

    def _clear(self):
        return self.__dict__.clear()

    def _get(self, *args):
        return self.__dict__.get(*args)

    def _items(self):
        return self.__dict__.items()

    def _keys(self):
        return self.__dict__.keys()

    def _pop(self, *args):
        return self.__dict__.pop(*args)

    def _setdefault(self, *args):
        return self.__dict__.setdefault(*args)


class Context(DataObject):

    _fields = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._fields:
            unknowns = [name for name in self._keys() if name not in self._fields]
            if unknowns:
                raise ValueError(f"Unexpected field names: {unknowns!r}")

    @classmethod
    def _new(cls, *args, **kwargs):
        if len(args) > len(cls._fields):
            raise TypeError("Too many positional arguments")
        return cls(zip(cls._fields, args), **kwargs)

    @classmethod
    def _make(cls, obj):
        if obj is None:
            return cls()
        elif isinstance(obj, cls):
            return obj
        else:
            return cls(obj)

    def __setattr__(self, name, value):
        if name in self:
            raise TypeError(f"Field '{name}' already set")
        super().__setattr__(name, value)

    def _branch(self, **kwargs):
        if not kwargs:
            return self
        items = dict(
            ((k, v) for k, v in kwargs.items() if v is not None and v != self[k])
        )
        if items:
            return self.__class__(self, **items)
        else:
            return self

    def _setdefaults(self, source):
        if not isinstance(source, dict):
            source = source.__dict__
        new_values = source.copy()
        new_values.update(self.__dict__)
        self.__dict__.update(new_values)
        return self

    def __bool__(self):
        return True

    __nonzero__ = __bool__


class FrozenDict(Mapping):
    def __init__(self, value):
        self._value = dict(value)

    def __getitem__(self, key):
        return self._value[key]

    def __iter__(self):
        return iter(self._value)

    def __len__(self):
        return len(self._value)

    def __hash__(self):
        try:
            return self._hash
        except AttributeError:
            self._hash = 0
            for k, v in self._value.items():
                self._hash ^= hash(k)
                self._hash ^= hash(v)

            return self._hash

    def __repr__(self):
        return repr(self._value)

    def __str__(self):
        return str(self._value)


class FrozenList(Sequence):
    def __init__(self, value):
        self._list = list(value)

    def __getitem__(self, index):
        return self._list[index]

    def __len__(self):
        return len(self._list)

    def __hash__(self):
        try:
            return self._hash
        except AttributeError:
            self._hash = 0
            for e in self._list:
                self._hash ^= hash(e)
            return self._hash

    def __repr__(self):
        return repr(self._list)

    def __str__(self):
        return str(self._list)

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        for i in range(len(self)):
            if self[i] != other[i]:
                return False
        return True



import sys
from collections.abc import Sequence

try:
    from _thread import get_ident
except ImportError:
    from _dummy_thread import get_ident

__all__ = [
    "get_ident",
    "setdefault",
    "Constant",
    "listify",
    "get_all_subclasses",
    "ImportStringError",
    "import_string",
]


def setdefault(obj, attr, value, search_mro=False, overwrite_none=False):
    if search_mro:
        exists = hasattr(obj, attr)
    else:
        exists = attr in obj.__dict__
    if exists and overwrite_none:
        if getattr(obj, attr) is None:
            exists = False
    if exists:
        value = getattr(obj, attr)
    else:
        setattr(obj, attr, value)
    return value


class Constant(int):
    def __new__(cls, name, value):
        return int.__new__(cls, value)

    def __init__(self, name, value):
        self.name = name
        int.__init__(self)

    def __repr__(self):
        return self.name

    __str__ = __repr__


def listify(value):
    if isinstance(value, list):
        return value
    if value is None:
        return []
    if isinstance(value, str):
        return [value]
    if isinstance(value, Sequence):
        return list(value)
    return [value]


def get_all_subclasses(cls):
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


class ImportStringError(ImportError):

    """Provides information about a failed :func:`import_string` attempt.

    Code taken from werzeug BSD license at https://github.com/pallets/werkzeug/blob/master/LICENSE
    """

    #: String in dotted notation that failed to be imported.
    import_name = None
    #: Wrapped exception.
    exception = None

    def __init__(self, import_name, exception):
        self.import_name = import_name
        self.exception = exception

        msg = (
            "import_string() failed for %r. Possible reasons are:\n\n"
            "- missing __init__.py in a package;\n"
            "- package or module path not included in sys.path;\n"
            "- duplicated package or module name taking precedence in "
            "sys.path;\n"
            "- missing module, class, function or variable;\n\n"
            "Debugged import:\n\n%s\n\n"
            "Original exception:\n\n%s: %s"
        )

        name = ""
        tracked = []
        for part in import_name.replace(":", ".").split("."):
            name += (name and ".") + part
            imported = import_string(name, silent=True)
            if imported:
                tracked.append((name, getattr(imported, "__file__", None)))
            else:
                track = [f"- {n!r} found in {i!r}." for n, i in tracked]
                track.append(f"- {name!r} not found.")
                msg = msg % (
                    import_name,
                    "\n".join(track),
                    exception.__class__.__name__,
                    str(exception),
                )
                break

        ImportError.__init__(self, msg)

    def __repr__(self):
        return f"<{self.__class__.__name__}({self.import_name!r}, {self.exception!r}).>"


def import_string(import_name, silent=False):
    """Imports an object based on a string.  This is useful if you want to
    use import paths as endpoints or something similar.  An import path can
    be specified either in dotted notation (``xml.sax.saxutils.escape``)
    or with a colon as object delimiter (``xml.sax.saxutils:escape``).

    If `silent` is True the return value will be `None` if the import fails.

    Code taken from werzeug BSD license at https://github.com/pallets/werkzeug/blob/master/LICENSE

    :param import_name: the dotted name for the object to import.
    :param silent: if set to `True` import errors are ignored and
                   `None` is returned instead.
    :return: imported object
    """
    # force the import name to automatically convert to strings
    # __import__ is not able to handle unicode strings in the fromlist
    # if the module is a package
    import_name = str(import_name).replace(":", ".")
    try:
        try:
            __import__(import_name)
        except ImportError:
            if "." not in import_name:
                raise
        else:
            return sys.modules[import_name]

        module_name, obj_name = import_name.rsplit(".", 1)
        try:
            module = __import__(module_name, None, None, [obj_name])
        except ImportError:
            # support importing modules not yet set up by the parent module
            # (or package for that matter)
            module = import_string(module_name)

        try:
            return getattr(module, obj_name)
        except AttributeError as e:
            raise ImportError(e) from e

    except ImportError as e:
        if not silent:
            raise ImportStringError(import_name, e) from e



import datetime
import random
import sys
import uuid
from decimal import Decimal
from fractions import Fraction

import pytest

from schematics.datastructures import Context
from schematics.exceptions import ConversionError, DataError, ValidationError
from schematics.models import Model
from schematics.types import *
from schematics.types.base import get_range_endpoints
from schematics.types.compound import *

try:
    from schematics.compat import bytes, str
except ImportError:
    pass


_uuid = uuid.UUID("3ce85e48-3028-409c-a07c-c8ee3d16d5c4")


def test_type_repr():
    class Bar(Model):
        pass

    class Foo(Model):
        intfield = IntType()
        modelfield = ModelType(Bar)
        listfield = ListType(ListType(StringType))

    assert repr(Foo.intfield) == "<IntType() instance on Foo as 'intfield'>"
    assert repr(Foo.listfield) == "<ListType(ListType) instance on Foo as 'listfield'>"
    assert repr(Foo.listfield.field.field) == "<StringType() instance on Foo>"
    assert repr(Foo.modelfield) == "<ModelType(Bar) instance on Foo as 'modelfield'>"
    assert repr(DateTimeType()) == "<DateTimeType() instance>"


@pytest.mark.parametrize("choice_type", [list, tuple, set, frozenset])
def test_choices(choice_type):
    assert BaseType(choices=choice_type("foo"))


def test_string_choices():
    with pytest.raises(TypeError):
        BaseType(choices="foo")


def test_date():
    today = datetime.date(2013, 3, 1)

    date_type = DateType()

    with pytest.raises(ConversionError):
        date_type.to_native("2013/03/01")

    assert date_type("2013-03-01") == today

    assert date_type.to_primitive(today) == "2013-03-01"

    assert date_type.to_native(today) is today

    assert date_type.to_native(datetime.datetime(2013, 3, 1)) == today

    assert type(date_type._mock()) is datetime.date


def test_date_formats():
    today = datetime.date(2013, 3, 1)

    date_type = DateType(formats="%d.%m.%Y")

    with pytest.raises(ConversionError):
        date_type.to_native("2013-03-01")

    assert date_type("1.3.2013") == today

    assert date_type.to_primitive(today) == "2013-03-01"

    date_type = DateType(formats=("%d.%m.%Y", "%d/%m/%Y"))

    with pytest.raises(ConversionError):
        date_type.to_native("2013-03-01")

    assert date_type("1.3.2013") == today
    assert date_type("1/3/2013") == today

    assert date_type.to_primitive(today) == "2013-03-01"


def test_datetime():
    dt = datetime.datetime.now()
    assert DateTimeType()(dt.isoformat()) == dt

    obj = DateTimeType(formats="foo")
    assert obj.formats == ["foo"]


def test_datetime_format():
    dt_input = "2013.03.07 15:31"
    dt = datetime.datetime(2013, 3, 7, 15, 31)
    assert DateTimeType(formats=["%Y.%m.%d %H:%M"])(dt_input) == dt


def test_datetime_to_native():
    with pytest.raises(ConversionError):
        DateTimeType().to_native("foo")


def test_datetime_primitive():
    output = "2013.03.07 15:31"
    dt = datetime.datetime(2013, 3, 7, 15, 31)
    dt_type = DateTimeType(serialized_format="%Y.%m.%d %H:%M")
    assert dt_type(dt.isoformat()) == dt
    assert dt_type.to_primitive(dt) == output

    obj = DateTimeType(serialized_format=str)
    assert obj.to_primitive(123) == "123"


def test_datetime_accepts_datetime():
    output = "2013.03.07 15:31"
    dt = datetime.datetime(2013, 3, 7, 15, 31)
    dt_type = DateTimeType(serialized_format="%Y.%m.%d %H:%M")
    assert dt_type(dt) == dt
    assert dt_type.to_primitive(dt) == output


def test_timedelta():
    type_ = TimedeltaType()
    delta = datetime.timedelta(seconds=10)
    assert type_(delta.total_seconds()) == delta


@pytest.mark.parametrize(
    "value",
    (
        42.42,
        "42.42",
        10000,
        0,
    ),
)
def test_timedelta_to_native_valid_conversion(value):
    type_ = TimedeltaType()
    assert type_.to_native(value) == datetime.timedelta(seconds=float(value))


@pytest.mark.parametrize(
    "value",
    (
        "foo",
        [],
    ),
)
def test_timedelta_to_native_invalid_conversion(value):
    type_ = TimedeltaType()
    with pytest.raises(ConversionError):
        type_.to_native(value)


@pytest.mark.parametrize(
    "value",
    (
        42.42,
        "42.42",
        10000,
        0,
    ),
)
def test_timedelta_to_primitive_valid_serialization(value):
    type_ = TimedeltaType()
    td = datetime.timedelta(seconds=float(value))
    assert type_.to_primitive(td) == int(float(value))


@pytest.mark.parametrize(
    "precision",
    (
        TimedeltaType.DAYS,
        TimedeltaType.SECONDS,
        TimedeltaType.MICROSECONDS,
        TimedeltaType.MILLISECONDS,
        TimedeltaType.MINUTES,
        TimedeltaType.HOURS,
        TimedeltaType.WEEKS,
    ),
)
def test_timedelta_to_primitive_valid_precision(precision):
    type_ = TimedeltaType(precision=precision)
    value = 42
    td = datetime.timedelta(**{precision: float(value)})
    assert type_.to_primitive(td) == int(float(value))


def test_timedelta_mock():
    type_ = TimedeltaType()
    mocked = type_._mock()
    assert isinstance(mocked, datetime.timedelta)


def test_int():
    intfield = IntType()
    i = lambda x: (intfield(x), type(intfield(x)))
    with pytest.raises(ConversionError):
        i("3.0")
    # from decimal
    with pytest.raises(ConversionError):
        i(Decimal("3.2"))
    # from fraction
    with pytest.raises(ConversionError):
        i(Fraction(7, 2))
    # from uuid
    with pytest.raises(ConversionError):
        i(_uuid)


def test_int_strict():
    intfield = IntType(strict=True)
    i = lambda x: (intfield(x), type(intfield(x)))
    # from int
    assert i(3) == (3, int)
    # from bool
    assert i(True) == (1, int)
    # from float
    with pytest.raises(ConversionError):
        i(3.0)
    # from string
    assert i("3") == (3, int)
    with pytest.raises(ConversionError):
        i("3.0")
    # from decimal
    with pytest.raises(ConversionError):
        assert i(Decimal("3")) == (3, int)
    with pytest.raises(ConversionError):
        assert i(Decimal("3.0")) == (3, int)
    # from fraction
    with pytest.raises(ConversionError):
        assert i(Fraction(30, 10)) == (3, int)
    # from uuid
    with pytest.raises(ConversionError):
        i(_uuid)


@pytest.mark.parametrize(
    "field_type,raw,expected",
    [
        (DecimalType, 0.3, Decimal("0.3")),
        (DecimalType, Decimal("0." + 10 * "9"), Decimal("0." + 10 * "9")),
        (DecimalType, False, Decimal(0)),
        (FloatType, "-1.3", -1.3),
        (FloatType, "3", 3.0),
        (FloatType, "3.2", 3.2),
        (FloatType, "3.210987e6", 3210987.0),
        (FloatType, 3, 3.0),
        (FloatType, 3.2, 3.2),
        (FloatType, Decimal("3"), 3.0),
        (FloatType, Decimal("3.2"), 3.2),
        (FloatType, Fraction(7, 2), 3.5),
        (FloatType, True, 1.0),
        (IntType, "32", 32),
        (IntType, 0.0, 0),
        (IntType, 3, 3),
        (IntType, Decimal("3"), 3),
        (IntType, Decimal("3.0"), 3),
        (IntType, False, 0),
        (IntType, Fraction(30, 10), 3),
        (IntType, True, 1),
    ],
)
def test_number_to_native(field_type, raw, expected):
    field = field_type()
    got = field.to_native(raw)
    assert (got, type(got)) == (expected, type(expected))


@pytest.mark.parametrize(
    "field_type,raw,error_msg",
    [
        (DecimalType, "a", "Value 'a' is not decimal."),
        (FloatType, "abc", "Value 'abc' is not float."),
        (FloatType, _uuid, "Value '{}' is not float.".format(_uuid)),
        (IntType, "3.0", "Value '3.0' is not int."),
        (IntType, 3.2, "Value '3.2' is not int."),
        (IntType, _uuid, "Value '{}' is not int.".format(_uuid)),
    ],
)
def test_number_conversion_error(field_type, raw, error_msg):
    with pytest.raises(ConversionError) as exc_info:
        field_type()(raw)
    assert exc_info.value.to_primitive() == [error_msg]


@pytest.mark.parametrize(
    "field_type,min_value,max_value",
    [
        (DecimalType, 1.3, 3.7),
        (FloatType, -1.3, 4.5),
        (IntType, 0, 10),
    ],
)
def test_number_mock(field_type, min_value, max_value):
    field = field_type(required=True, min_value=min_value, max_value=max_value)
    mock = field.mock()
    assert type(mock) is field_type.native_type
    field.validate(mock)


@pytest.mark.parametrize(
    "field_type,value,primitive",
    [
        (DecimalType, Decimal("1.3"), "1.3"),
        (FloatType, 1.3, 1.3),
        (IntType, 13, 13),
    ],
)
def test_number_to_primitive(field_type, value, primitive):
    got = field_type().to_primitive(value)
    assert (got, type(got)) == (primitive, type(primitive))


def test_int_validation():
    with pytest.raises(ConversionError):
        IntType().validate("foo")
    assert IntType().validate(5001) == 5001


def test_custom_validation_functions():
    class UppercaseType(BaseType):
        def validate_uppercase(self, value, context=None):
            if value.upper() != value:
                raise ValidationError("Value must be uppercase!")

        def validate_without_context(self, value):
            pass

    field = UppercaseType()

    field.validate("UPPERCASE")

    with pytest.raises(ValidationError):
        field.validate("lowercase")


def test_custom_validation_function_and_inheritance():
    class UppercaseType(StringType):
        def validate_uppercase(self, value, context=None):
            if value.upper() != value:
                raise ValidationError("Value must be uppercase!")

    class MUppercaseType(UppercaseType):
        def __init__(self, number_of_m_chars, **kwargs):
            self.number_of_m_chars = number_of_m_chars

            super().__init__(**kwargs)

        def validate_contains_m_chars(self, value, context=None):
            if value.count("M") != self.number_of_m_chars:
                raise ValidationError(
                    f"Value must contain {self.number_of_m_chars} 'm' characters"
                )

    field = MUppercaseType(number_of_m_chars=3)

    field.validate("MMM")

    with pytest.raises(ValidationError):
        field.validate("mmm")

    with pytest.raises(ValidationError):
        field.validate("MM")


def test_string_type_required():
    class M(Model):
        field = StringType(required=True)

    with pytest.raises(DataError):
        M({"field": None}).validate()


def test_string_type_accepts_none():
    class M(Model):
        field = StringType()

    M({"field": None}).validate()


def test_string_required_accepts_empty_string():
    class M(Model):
        field = StringType()

    M({"field": ""}).validate()


def test_string_min_length_doesnt_accept_empty_string():
    field = StringType(min_length=1)
    with pytest.raises(ValidationError):
        field.validate("")


def test_string_regex():
    StringType(regex=r"\d+").validate("1")

    with pytest.raises(ValidationError):
        StringType(regex=r"\d+").validate("a")


def test_string_to_native():
    field = StringType()

    with pytest.raises(ConversionError):
        field.to_native(3.14)
    with pytest.raises(ConversionError):
        field.to_native(b"\xE0\xA0")  # invalid UTF-8 sequence
    with pytest.raises(ConversionError):
        field.to_native(True)

    if sys.version_info[0] == 2:
        strings = [
            ("abcdefg", "abcdefg"),
            ("abcdefg", "abcdefg"),
            ("abc éíçßµ", "abc éíçßµ"),
            ("\xC3\xA4", "ä"),
            (12345, "12345"),
        ]
    else:
        strings = [
            ("abcdefg", "abcdefg"),
            (b"abcdefg", "abcdefg"),
            ("abc éíçßµ", "abc éíçßµ"),
            (b"\xC3\xA4", "ä"),
            (12345, "12345"),
        ]

    for input, expected in strings:
        res = field.to_native(input)
        assert res == expected
        assert type(res) is str


def test_string_max_length_is_enforced():
    with pytest.raises(ValidationError):
        StringType(max_length=2).validate_length("foo")


def test_multilingualstring_should_only_take_certain_types():
    mls = MultilingualStringType()

    mls(None)
    mls({})

    with pytest.raises(ConversionError):
        mls(123)

    with pytest.raises(ConversionError):
        mls([])

    with pytest.raises(ConversionError):
        mls("foo")


def test_multilingualstring_should_validate_length():
    MultilingualStringType(min_length=3).validate({"en_US": "foo"})
    MultilingualStringType(max_length=3).validate({"en_US": "foo"})

    with pytest.raises(ValidationError):
        MultilingualStringType(min_length=4).validate({"en_US": "foo"})

    with pytest.raises(ValidationError):
        MultilingualStringType(max_length=2).validate({"en_US": "foo"})


def test_multilingualstring_should_validate_regex():
    MultilingualStringType(regex="^[a-z]*$").validate({"en_US": "foo"})

    with pytest.raises(ValidationError):
        MultilingualStringType(regex="^[a-z]*$").validate({"en_US": "123"})

    with pytest.raises(ValidationError):
        MultilingualStringType(locale_regex=r"^\d*$").validate({"en_US": "foo"})

    assert (
        MultilingualStringType(locale_regex=None).validate_regex({"en_US": "foo"})
        is None
    )


def test_multilingualstring_should_handle_none():
    mls = MultilingualStringType(default_locale="en_US")

    assert mls.to_primitive(None) is None


def test_multilingualstring_should_handle_castable_values():
    mls = MultilingualStringType(default_locale="en_US")

    assert mls.to_primitive({"en_US": 123}) == "123"


def test_multilingualstring_should_enforce_noncastable_values():
    mls = MultilingualStringType(default_locale="en_US")

    with pytest.raises(ConversionError):
        mls.to_primitive({"en_US": 123.0}) == "123"


def test_multilingual_string_should_emit_string_with_default_locale():
    mls = MultilingualStringType(default_locale="en_US")

    assert mls.to_primitive({"en_US": "snake", "fr_FR": "serpent"}) == "snake"


def test_multilingual_string_should_emit_string_with_explicit_locale():
    mls = MultilingualStringType(default_locale="en_US")

    assert (
        mls.to_primitive(
            {"en_US": "snake", "fr_FR": "serpent"},
            context=Context(app_data={"locale": "fr_FR"}),
        )
        == "serpent"
    )


def test_multilingual_string_should_require_a_locale():
    mls = MultilingualStringType()

    with pytest.raises(ConversionError):
        mls.to_primitive({"foo": "bar"})


def test_multilingual_string_without_matching_locale_should_explode():
    mls = MultilingualStringType()

    with pytest.raises(ConversionError):
        mls.to_primitive({"fr_FR": "serpent"})

    with pytest.raises(ConversionError):
        mls.to_primitive(
            {"en_US": "snake"}, context=Context(app_data={"locale": "fr_FR"})
        )


def test_multilingual_string_should_accept_lists_of_locales():
    strings = {
        "en_US": "snake",
        "fr_FR": "serpent",
        "es_MX": "serpiente",
    }

    mls = MultilingualStringType(default_locale=["foo", "fr_FR", "es_MX"])

    assert mls.to_primitive(strings) == "serpent"
    assert (
        mls.to_primitive(
            strings, context=Context(app_data={"locale": ["es_MX", "bar"]})
        )
        == "serpiente"
    )

    mls = MultilingualStringType()

    assert (
        mls.to_primitive(
            strings, context=Context(app_data={"locale": ["foo", "es_MX", "fr_FR"]})
        )
        == "serpiente"
    )


def test_boolean_to_native():
    bool_field = BooleanType()

    for true_value in ["True", "1", 1, True]:
        assert bool_field.to_native(true_value)

    for false_value in ["False", "0", 0, False]:
        assert not bool_field.to_native(false_value)

    for bad_value in ["TrUe", "foo", 2, None, 1.0]:
        with pytest.raises(ConversionError):
            bool_field.to_native(bad_value)

    assert type(BooleanType(required=True).mock()) is bool


def test_geopoint_mock():
    geo = GeoPointType(required=True)
    geo_point = geo.mock()
    assert geo_point[0] >= -90
    assert geo_point[0] <= 90
    assert geo_point[1] >= -180
    assert geo_point[1] <= 180


def test_geopoint_to_native():
    geo = GeoPointType(required=True)

    with pytest.raises(ConversionError):
        native = geo.to_native((10,))

    with pytest.raises(ConversionError):
        native = geo.to_native({"1": "-20", "2": "18"})

    with pytest.raises(ConversionError):
        native = geo.to_native(["-20", "18"])

    with pytest.raises(ConversionError):
        native = geo.to_native("-20, 18")

    class Point:
        def __len__(self):
            return 2

    with pytest.raises(ConversionError):
        native = geo.to_native(Point())

    native = geo.to_native([89, -12])
    assert native == [89, -12]

    latitude = random.uniform(-90, 90)
    longitude = random.uniform(-180, 180)
    point = [latitude, longitude]

    native = geo.to_native(point)
    assert native == point


def test_geopoint_range():
    geo = GeoPointType(required=True)

    with pytest.raises(ValidationError) as ve:
        geo.validate((-91, 110))

    with pytest.raises(ValidationError) as ve:
        geo.validate((90.12345, 65))

    with pytest.raises(ValidationError) as ve:
        geo.validate((23, -181))

    with pytest.raises(ValidationError) as ve:
        geo.validate((28.2342323, 181.123141))


def test_metadata():
    metadata = {"description": "amazing", "extra": {"abc": 1}}
    assert StringType(metadata=metadata).metadata == metadata
    assert StringType().metadata == {}



"""Schematics models."""

import inspect
from collections import ChainMap, OrderedDict
from copy import deepcopy
from types import FunctionType, MappingProxyType
from typing import List

from . import schema
from .datastructures import Context
from .exceptions import (
    DataError,
    MockCreationError,
    UndefinedValueError,
    UnknownFieldError,
)
from .iteration import atoms
from .transforms import convert, export_loop, to_native, to_primitive
from .types.base import BaseType
from .types.serializable import Serializable
from .undefined import Undefined
from .util import get_ident
from .validate import prepare_validator, validate

__all__: List[str] = []


class FieldDescriptor:
    """
    ``FieldDescriptor`` instances serve as field accessors on models.
    """

    def __init__(self, name):
        """
        :param name:
            The field's name
        """
        self.name = name

    def __get__(self, instance, cls):
        """
        For a model instance, returns the field's current value.
        For a model class, returns the field's type object.
        """
        if instance is None:
            return cls._schema.fields[self.name]
        value = instance._data.get(self.name, Undefined)
        if value is Undefined:
            raise UndefinedValueError(instance, self.name)
        return value

    def __set__(self, instance, value):
        """
        Sets the field's value.
        """
        field = instance._schema.fields[self.name]
        value = field.pre_setattr(value)
        instance._data.converted[self.name] = value

    def __delete__(self, instance):
        """
        Deletes the field's value.
        """
        del instance._data[self.name]


class ModelMeta(type):
    """
    Metaclass for Models.
    """

    def __new__(mcs, name, bases, attrs):
        """
        This metaclass parses the declarative Model into a corresponding Schema,
        then adding it as the `_schema` attribute to the host class.
        """

        # Structures used to accumulate meta info
        fields = OrderedDict()
        validator_functions = {}  # Model level
        options_members = {}

        # Accumulate metas info from parent classes
        for base in reversed(bases):
            if hasattr(base, "_schema"):
                fields.update(deepcopy(base._schema.fields))
                options_members.update(dict(base._schema.options))
                validator_functions.update(base._schema.validators)

        # Parse this class's attributes into schema structures
        for key, value in attrs.items():
            if key.startswith("validate_") and isinstance(
                value, (FunctionType, classmethod)
            ):
                validator_functions[key[9:]] = prepare_validator(value, 4)
            if isinstance(value, BaseType):
                fields[key] = value
            elif isinstance(value, Serializable):
                fields[key] = value

        # Convert declared fields into descriptors for new class
        fields = OrderedDict(
            sorted(
                (kv for kv in fields.items()),
                key=lambda i: i[1]._position_hint,
            )
        )
        for key, field in fields.items():
            if isinstance(field, BaseType):
                attrs[key] = FieldDescriptor(key)
            elif isinstance(field, Serializable):
                attrs[key] = field

        klass = type.__new__(mcs, name, bases, attrs)

        # Parse schema options
        options = mcs._read_options(name, bases, attrs, options_members)

        # Parse meta data into new schema
        klass._schema = schema.Schema(
            name,
            model=klass,
            options=options,
            validators=validator_functions,
            *(schema.Field(k, t) for k, t in fields.items()),
        )

        return klass

    @classmethod
    def _read_options(mcs, name, bases, attrs, options_members):
        """
        Parses model `Options` class into a `SchemaOptions` instance.
        """
        options_class = attrs.get("__optionsclass__", schema.SchemaOptions)
        if "Options" in attrs:
            for key, value in inspect.getmembers(attrs["Options"]):
                if key.startswith("__"):
                    continue
                if key.startswith("_"):
                    extras = options_members.get("extras", {}).copy()
                    extras.update({key: value})
                    options_members["extras"] = extras
                elif key == "roles":
                    roles = options_members.get("roles", {}).copy()
                    roles.update(value)
                    options_members[key] = roles
                else:
                    options_members[key] = value
        return options_class(**options_members)


class ModelDict(ChainMap):

    __slots__ = ["_unsafe", "_converted", "__valid", "_valid"]

    def __init__(self, unsafe=None, converted=None, valid=None):
        self._unsafe = unsafe if unsafe is not None else {}
        self._converted = converted if converted is not None else {}
        self.__valid = valid if valid is not None else {}
        self._valid = MappingProxyType(self.__valid)
        super().__init__(self._unsafe, self._converted, self._valid)

    @property
    def unsafe(self):
        return self._unsafe

    @unsafe.setter
    def unsafe(self, value):
        self._unsafe = value
        self.maps[0] = self._unsafe

    @property
    def converted(self):
        return self._converted

    @converted.setter
    def converted(self, value):
        self._converted = value
        self.maps[1] = self._converted

    @property
    def valid(self):
        return self._valid

    @valid.setter
    def valid(self, value):
        self._valid = MappingProxyType(value)
        self.maps[2] = self._valid

    def __delitem__(self, key):
        did_delete = False
        for data in [self.__valid, self._converted, self._unsafe]:
            try:
                del data[key]
                did_delete = True
            except KeyError:
                pass
        if not did_delete:
            raise KeyError(key)

    def __repr__(self):
        return repr(dict(self))


class Model(metaclass=ModelMeta):

    """
    Enclosure for fields and validation. Same pattern deployed by Django
    models, SQLAlchemy declarative extension and other developer friendly
    libraries.

    :param Mapping raw_data:
        The data to be imported into the model instance.
    :param Mapping deserialize_mapping:
        Can be used to provide alternative input names for fields. Values may be
        strings or lists of strings, keyed by the actual field name.
    :param bool partial:
        Allow partial data to validate. Essentially drops the ``required=True``
        settings from field definitions. Default: True
    :param bool strict:
        Complain about unrecognized keys. Default: True
    """

    def __init__(
        self,
        raw_data=None,
        trusted_data=None,
        deserialize_mapping=None,
        init=True,
        partial=True,
        strict=True,
        validate=False,
        app_data=None,
        lazy=False,
        **kwargs,
    ):
        kwargs.setdefault("init_values", init)
        kwargs.setdefault("apply_defaults", init)

        if lazy:
            self._data = ModelDict(unsafe=raw_data, valid=trusted_data)
            return

        self._data = ModelDict(valid=trusted_data)
        data = self._convert(
            raw_data,
            trusted_data=trusted_data,
            mapping=deserialize_mapping,
            partial=partial,
            strict=strict,
            validate=validate,
            new=True,
            app_data=app_data,
            **kwargs,
        )
        self._data.converted = data
        if validate:
            self.validate(partial=partial, app_data=app_data, **kwargs)

    def validate(self, partial=False, convert=True, app_data=None, **kwargs):
        """
        Validates the state of the model. If the data is invalid, raises a ``DataError``
        with error messages.

        :param bool partial:
            Allow partial data to validate. Essentially drops the ``required=True``
            settings from field definitions. Default: False
        :param convert:
            Controls whether to perform import conversion before validating.
            Can be turned off to skip an unnecessary conversion step if all values
            are known to have the right datatypes (e.g., when validating immediately
            after the initial import). Default: True
        """
        if not self._data.converted and partial:
            return  # no new input data to validate
        try:
            data = self._convert(
                validate=True,
                partial=partial,
                convert=convert,
                app_data=app_data,
                **kwargs,
            )
            self._data.valid = data
        except DataError as e:
            valid = dict(self._data.valid)
            valid.update(e.partial_data)
            self._data.valid = valid
            raise
        finally:
            self._data.converted = {}

    def import_data(self, raw_data, recursive=False, **kwargs):
        """
        Converts and imports the raw data into an existing model instance.

        :param raw_data:
            The data to be imported.
        """
        data = self._convert(
            raw_data, trusted_data=dict(self), recursive=recursive, **kwargs
        )
        self._data.converted.update(data)
        if kwargs.get("validate"):
            self.validate(convert=False)
        return self

    def _convert(self, raw_data=None, context=None, **kwargs):
        """
        Converts the instance raw data into richer Python constructs according
        to the fields on the model, validating data if requested.

        :param raw_data:
            New data to be imported and converted
        """
        raw_data = (
            {key: raw_data[key] for key in raw_data}
            if raw_data
            else self._data.converted
        )
        kwargs["trusted_data"] = kwargs.get("trusted_data") or {}
        kwargs["convert"] = getattr(context, "convert", kwargs.get("convert", True))
        if self._data.unsafe:
            self._data.unsafe.update(raw_data)
            raw_data = self._data.unsafe
            self._data.unsafe = {}
            kwargs["convert"] = True
        should_validate = getattr(context, "validate", kwargs.get("validate", False))
        func = validate if should_validate else convert
        return func(
            self._schema, self, raw_data=raw_data, oo=True, context=context, **kwargs
        )

    def export(self, field_converter=None, role=None, app_data=None, **kwargs):
        return export_loop(
            self._schema,
            self,
            field_converter=field_converter,
            role=role,
            app_data=app_data,
            **kwargs,
        )

    def to_native(self, role=None, app_data=None, **kwargs):
        return to_native(self._schema, self, role=role, app_data=app_data, **kwargs)

    def to_primitive(self, role=None, app_data=None, **kwargs):
        return to_primitive(self._schema, self, role=role, app_data=app_data, **kwargs)

    def serialize(self, *args, **kwargs):
        raw_data = self._data.converted
        try:
            self.validate(apply_defaults=True)
        except DataError:
            pass
        data = self.to_primitive(*args, **kwargs)
        self._data.converted = raw_data
        return data

    def atoms(self):
        """
        Iterator for the atomic components of a model definition and relevant
        data that creates a 3-tuple of the field's name, its type instance and
        its value.
        """
        return atoms(self._schema, self)

    def __iter__(self):
        return (
            k
            for k in self._schema.fields
            if k in self._data and getattr(self._schema.fields[k], "fset", None) is None
        )

    def keys(self):
        return list(iter(self))

    def items(self):
        return [(k, self._data[k]) for k in self]

    def values(self):
        return [self._data[k] for k in self]

    def get(self, key, default=None):
        return getattr(self, key, default)

    @classmethod
    def _append_field(cls, field_name, field_type):
        """
        Add a new field to this class.

        :type field_name: str
        :param field_name:
            The name of the field to add.
        :type field_type: BaseType
        :param field_type:
            The type to use for the field.
        """
        cls._schema.append_field(schema.Field(field_name, field_type))
        setattr(cls, field_name, FieldDescriptor(field_name))

    @classmethod
    def get_mock_object(cls, context=None, overrides=None):
        """Get a mock object.

        :param dict context:
        :param dict overrides: overrides for the model
        """
        context = Context._make(context)
        context._setdefault("memo", set())
        context.memo.add(cls)
        values = {}
        overrides = overrides or {}
        for name, field in cls._schema.fields.items():
            if name in overrides:
                continue
            if getattr(field, "model_class", None) in context.memo:
                continue
            try:
                values[name] = field.mock(context)
            except MockCreationError as exc:
                raise MockCreationError(f"{name}: {exc.args[0]}") from exc
        values.update(overrides)
        return cls(values)

    def __getitem__(self, name):
        if name in self._schema.fields:
            return getattr(self, name)
        raise UnknownFieldError(self, name)

    def __setitem__(self, name, value):
        if name in self._schema.fields:
            return setattr(self, name, value)
        raise UnknownFieldError(self, name)

    def __delitem__(self, name):
        if name in self._schema.fields:
            return delattr(self, name)
        raise UnknownFieldError(self, name)

    def __contains__(self, name):
        serializables = {
            k for k, t in self._schema.fields.items() if isinstance(t, Serializable)
        }
        return (
            name in self._data and getattr(self, name, Undefined) is not Undefined
        ) or name in serializables

    def __len__(self):
        return len(self._data)

    def __eq__(self, other, memo=set()):
        if self is other:
            return True
        if type(self) is not type(other):
            return NotImplemented
        key = (id(self), id(other), get_ident())
        if key in memo:
            return True
        memo.add(key)
        try:
            for k in self:
                if self.get(k) != other.get(k):
                    return False
            return True
        finally:
            memo.remove(key)

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        model = self.__class__.__name__
        info = self._repr_info()
        if info:
            return f"<{model}: {info}>"
        return f"<{model} instance>"

    def _repr_info(self):
        """
        Subclasses may implement this method to augment the ``__repr__()`` output for the instance::

            class Person(Model):
                ...
                def _repr_info(self):
                    return self.name

            >>> Person({'name': 'Mr. Pink'})
            <Person: Mr. Pink>
        """
        return None



from schematics.models import Model
from schematics.types import IntType, StringType
from schematics.types.compound import DictType, ModelType
from schematics.types.serializable import serializable

try:
    long
except NameError:
    long = int


def test_basic_type():
    class PlayerInfo(Model):
        categories = DictType(StringType)

    info = PlayerInfo(
        dict(
            categories={
                "math": "math",
                "batman": "batman",
            }
        )
    )

    assert info.categories["math"] == "math"

    d = info.serialize()
    assert d == {
        "categories": {
            "math": "math",
            "batman": "batman",
        }
    }


def test_dict_type_with_model_type():
    class CategoryStats(Model):
        category_slug = StringType()
        total_wins = IntType()

    class PlayerInfo(Model):
        categories = DictType(ModelType(CategoryStats))
        # TODO: Maybe it would be cleaner to have
        #       DictType(CategoryStats) and implicitly convert to ModelType(CategoryStats)

    info = PlayerInfo(
        dict(
            categories={
                "math": {"category_slug": "math", "total_wins": 1},
                "batman": {"category_slug": "batman", "total_wins": 3},
            }
        )
    )

    math_stats = CategoryStats({"category_slug": "math", "total_wins": 1})
    assert info.categories["math"] == math_stats

    d = info.serialize()
    assert d == {
        "categories": {
            "math": {"category_slug": "math", "total_wins": 1},
            "batman": {"category_slug": "batman", "total_wins": 3},
        }
    }


def test_dict_type_with_model_type_init_with_instance():
    class ExperienceLevel(Model):
        level = IntType()

    class CategoryStats(Model):
        category_slug = StringType()
        total_wins = IntType()

        @serializable(type=ModelType(ExperienceLevel))
        def xp_level(self):
            return ExperienceLevel(dict(level=self.total_wins))

    class PlayerInfo(Model):
        id = IntType()
        categories = DictType(ModelType(CategoryStats))
        # TODO: Maybe it would be cleaner to have
        #       DictType(CategoryStats) and implicitly convert to ModelType(CategoryStats)

    math_stats = CategoryStats({"category_slug": "math", "total_wins": 1})

    info = PlayerInfo(
        dict(
            id=1,
            categories={
                "math": math_stats,
            },
        )
    )

    assert info.categories["math"] == math_stats

    d = info.serialize()

    assert d == {
        "id": 1,
        "categories": {
            "math": {
                "category_slug": "math",
                "total_wins": 1,
                "xp_level": {"level": 1},
            },
        },
    }


def test_with_empty():
    class CategoryStatsInfo(Model):
        slug = StringType()

    class PlayerInfo(Model):
        categories = DictType(
            ModelType(CategoryStatsInfo),
            default=lambda: {},
            serialize_when_none=True,
        )

    info = PlayerInfo()

    assert info.categories == {}

    d = info.serialize()
    assert d == {
        "categories": {},
    }


def test_key_type():
    def player_id(value):
        return long(value)

    class CategoryStatsInfo(Model):
        slug = StringType()

    class PlayerInfo(Model):
        categories = DictType(ModelType(CategoryStatsInfo), coerce_key=player_id)

    stats = CategoryStatsInfo(
        {
            "slug": "math",
        }
    )

    info = PlayerInfo(
        {
            "categories": {1: {"slug": "math"}},
        }
    )

    assert info.categories == {1: stats}

    d = info.serialize()
    assert d == {"categories": {1: {"slug": "math"}}}



import pytest

from schematics.models import DataError, Model
from schematics.types import StringType
from schematics.types.compound import ListType, PolyModelType
from schematics.util import get_all_subclasses


class A(Model):  # fallback model (doesn't define a claim method)
    stringA = StringType()


class Aaa(A):
    pass


class B(A):
    stringB = StringType()
    regexp_test = StringType(regex="^[0-9]$", required=False)

    @classmethod
    def _claim_polymorphic(cls, data):
        return data.get("stringB") == "bbb"


class C(B):
    stringC = StringType()

    @classmethod
    def _claim_polymorphic(cls, data):
        return data.get("stringC") == "ccc"


class X(Model):
    pass


def claim_func(field, data):
    if "stringB" in data and field.name == "cfn":
        return B
    if "stringC" in data and field.name == "cfn":
        return C
    else:
        return None


class Foo(Model):
    base = PolyModelType(A)  # accepts any subclass for import and export
    strict = PolyModelType([A, B])  # accepts [A, B] for import and export
    nfb = PolyModelType([B, C])  # no fallback since A not present
    cfn = PolyModelType([B, C], claim_function=claim_func)


def test_get_all_subclasses():
    assert A.__subclasses__() == [Aaa, B]
    assert B.__subclasses__() == [C]
    assert C.__subclasses__() == []

    assert get_all_subclasses(A) == [Aaa, B, C]


def test_inheritance_based_polymorphic():  # base

    foo = Foo({"base": {"stringB": "bbb"}})
    assert type(foo.base) is B

    foo = Foo({"base": {"stringC": "ccc"}})
    assert type(foo.base) is C

    foo = Foo({"base": {}})  # unrecognizable input => fall back to A
    assert type(foo.base) is A

    foo = Foo({"base": Aaa()})
    assert type(foo.base) is Aaa
    foo.validate()
    foo.to_primitive()


def test_enumerated_polymorphic():  # strict

    foo = Foo({"strict": {"stringB": "bbb"}})
    assert type(foo.strict) is B

    foo = Foo({"strict": {}})  # unrecognizable input => fall back to A
    assert type(foo.strict) is A

    foo = Foo({"strict": B()})
    assert type(foo.strict) is B
    foo.validate()
    foo.to_primitive()

    Foo.strict.allow_subclasses = True
    foo = Foo({"strict": Aaa()})
    Foo.strict.allow_subclasses = False


def test_external_claim_function():  # cfn

    foo = Foo({"cfn": {"stringB": "bbb", "stringC": "ccc"}}, strict=False)
    assert type(foo.cfn) is B


def test_multiple_matches():

    with pytest.raises(Exception):
        foo = Foo({"base": {"stringB": "bbb", "stringC": "ccc"}})


def test_no_fallback():  # nfb

    with pytest.raises(Exception):
        foo = Foo({"nfb": {}})


def test_refuse_unrelated_import():

    with pytest.raises(Exception):
        foo = Foo({"base": D()})

    with pytest.raises(Exception):
        foo = Foo({"strict": Aaa()})


def test_refuse_unrelated_export():

    with pytest.raises(Exception):
        foo = Foo()
        foo.base = X()
        foo.to_primitive()

    with pytest.raises(Exception):
        foo = Foo()
        foo.strict = Aaa()
        foo.to_primitive()


def test_specify_model_by_name():
    class M(Model):
        single = PolyModelType("M")
        multi = PolyModelType([A, "M", C])
        nested = ListType(ListType(PolyModelType("M")))

    assert M.single.is_allowed_model(M())
    assert M.multi.is_allowed_model(M())
    assert M.nested.field.field.is_allowed_model(M())


def test_validate_recursively():
    # Should validate
    foo = Foo({"strict": {"stringB": "bbb", "regexp_test": "1"}})
    foo.validate()

    foo = Foo({"strict": {"stringB": "bbb", "regexp_test": "a"}})
    with pytest.raises(DataError):
        foo.validate()



import random
import re
from typing import Optional
from urllib.error import URLError
from urllib.parse import quote as urlquote
from urllib.parse import urlunsplit
from urllib.request import urlopen

from ..exceptions import StopValidationError, ValidationError
from ..translator import _
from .base import StringType, fill_template

__all__ = [
    "IPAddressType",
    "IPv4Type",
    "IPv6Type",
    "MACAddressType",
    "URLType",
    "EmailType",
]


# Character ranges

HEX = "0-9A-F"
ALPHA = "A-Z"
ALPHANUM = "A-Z0-9"


# IP address patterns

IPV4_OCTET = "( 25[0-5] | 2[0-4][0-9] | [0-1]?[0-9]{1,2} )"
IPV4 = r"( ((%(oct)s\.){3} %(oct)s) )" % {"oct": IPV4_OCTET}

IPV6_H16 = "[%s]{1,4}" % HEX
IPV6_L32 = "(%(h16)s:%(h16)s|%(ipv4)s)" % {"h16": IPV6_H16, "ipv4": IPV4}
IPV6 = r"""(
                                    (%(h16)s:){6}%(l32)s  |
                                ::  (%(h16)s:){5}%(l32)s  |
    (               %(h16)s )?  ::  (%(h16)s:){4}%(l32)s  |
    ( (%(h16)s:){,1}%(h16)s )?  ::  (%(h16)s:){3}%(l32)s  |
    ( (%(h16)s:){,2}%(h16)s )?  ::  (%(h16)s:){2}%(l32)s  |
    ( (%(h16)s:){,3}%(h16)s )?  ::  (%(h16)s:){1}%(l32)s  |
    ( (%(h16)s:){,4}%(h16)s )?  ::               %(l32)s  |
    ( (%(h16)s:){,5}%(h16)s )?  ::               %(h16)s  |
    ( (%(h16)s:){,6}%(h16)s )?  :: )""" % {
    "h16": IPV6_H16,
    "l32": IPV6_L32,
}


class IPAddressType(StringType):
    """A field that stores a valid IPv4 or IPv6 address."""

    VERSION: Optional[str] = None
    REGEX = re.compile(rf"^{IPV4}|{IPV6}$", re.I + re.X)

    @classmethod
    def valid_ip(cls, value):
        return bool(cls.REGEX.match(value))

    def validate_(self, value, context=None):
        if not self.valid_ip(value):
            raise ValidationError(_("Invalid IP%s address") % (self.VERSION or ""))

    def _mock(self, context=None):
        return random.choice([IPv4Type, IPv6Type])(required=self.required).mock()


class IPv4Type(IPAddressType):
    """A field that stores a valid IPv4 address."""

    VERSION = "v4"
    REGEX = re.compile(rf"^{IPV4}$", re.I + re.X)

    def _mock(self, context=None):
        return ".".join(str(random.randrange(256)) for _ in range(4))


class IPv6Type(IPAddressType):
    """A field that stores a valid IPv6 address."""

    VERSION = "v6"
    REGEX = re.compile(rf"^{IPV6}$", re.I + re.X)

    def _mock(self, context=None):
        return "2001:db8:" + ":".join(
            "%x" % (random.randrange(1 << 16)) for _ in range(6)
        )


# MAC address


class MACAddressType(StringType):
    """A field that stores a valid MAC address."""

    REGEX = re.compile(
        r"""
                         (
                             ^([0-9a-f]{2}[-]){5}([0-9a-f]{2})$
                            |^([0-9a-f]{2}[:]){5}([0-9a-f]{2})$
                            |^([0-9a-f]{12})
                            |^([0-9a-f]{6}[-:]([0-9a-f]{6}))$
                            |^([0-9a-f]{4}(\.[0-9a-f]{4}){2})$
                         )
                         """,
        re.I + re.X,
    )

    def _mock(self, context=None):
        return ":".join("%02x" % (random.randrange(256)) for _ in range(6))

    def validate_(self, value, context=None):
        if not bool(self.REGEX.match(value)):
            raise ValidationError(_("Invalid MAC address"))

    def to_primitive(self, value, context=None):
        value = value.replace(":", "").replace(".", "").replace("-", "")
        return ":".join(value[i : i + 2] for i in range(0, len(value), 2))


# URI patterns

GEN_DELIMS = set(":/?#[]@")
SUB_DELIMS = set("!$&'()*+,;=")
UNRESERVED = set("-_.~")
PCHAR = SUB_DELIMS | UNRESERVED | set("%:@")
QUERY_EXTRAS = set("[]")  # nonstandard

VALID_CHARS = GEN_DELIMS | SUB_DELIMS | UNRESERVED | set("%")
VALID_CHAR_STRING = "".join(VALID_CHARS)
UNSAFE_CHAR_STRING = '\x00-\x20<>{}|"`\\^\x7F-\x9F'


def _chrcls(allowed_chars):
    """
    Given a subset of the URL-compatible special characters ``!#$%&'()*+,-./:;=?@[]_~``,
    returns a regex character class matching any URL-compatible character apart from the
    special characters not present in the provided set.
    """
    return (
        "^"
        + UNSAFE_CHAR_STRING
        + str.join("", VALID_CHARS - allowed_chars)
        .replace("%", "%%")
        .replace("]", r"\]")
        .replace("-", r"\-")
    )


URI_PATTERNS = {
    "scheme": r"[%s]+" % ("A-Z0-9.+-"),
    "user": r"[%s]+" % _chrcls(UNRESERVED | SUB_DELIMS | set("%:")),
    "port": r"[0-9]{2,5}",
    "host4": IPV4,
    "host6": r"[%s]+" % (HEX + ":"),
    "hostn": r"[%s]+" % _chrcls(set(".-")),
    "path": r"[%s]*" % _chrcls(PCHAR | set("/")),
    "query": r"[%s]*" % _chrcls(PCHAR | set("/?") | QUERY_EXTRAS),
    "frag": r"[%s]*" % _chrcls(PCHAR | set("/?")),
}


class URLType(StringType):

    """A field that validates the input as a URL.

    :param fqdn:
        if ``True`` the validation function will ensure hostname in URL
        is a Fully Qualified Domain Name.
    :param verify_exists:
        if ``True`` the validation function will make sure
        the URL is accessible (server responds with HTTP 2xx).
    """

    MESSAGES = {
        "invalid_url": "Not a well-formed URL.",
        "not_found": "URL could not be retrieved.",
    }

    URL_REGEX = re.compile(
        r"""^(
            (?P<scheme> %(scheme)s ) ://
        (   (?P<user>   %(user)s   ) @   )?
        (\[ (?P<host6>  %(host6)s  ) ]
          | (?P<host4>  %(host4)s  )
          | (?P<hostn>  %(hostn)s  )     )
        ( : (?P<port>   %(port)s   )     )?
            (?P<path> / %(path)s   )?
        (\? (?P<query>  %(query)s  )     )?
        (\# (?P<frag>   %(frag)s   )     )?)$
        """
        % URI_PATTERNS,
        re.I + re.X,
    )

    TLD_REGEX = re.compile(r"^( ([a-z]{2,}) | (xn--[a-z0-9]{4,}) )$", re.I + re.X)

    def __init__(self, fqdn=True, verify_exists=False, **kwargs):
        self.schemes = ["http", "https"]
        self.fqdn = fqdn
        self.verify_exists = verify_exists
        super().__init__(**kwargs)

    def _mock(self, context=None):
        return fill_template("http://a%s.ZZ", self.min_length, self.max_length)

    def valid_url(self, value):
        match = self.URL_REGEX.match(value)
        if not match:
            return False
        url = match.groupdict()

        if url["scheme"].lower() not in self.schemes:
            return False
        if url["host6"]:
            if IPv6Type.valid_ip(url["host6"]):
                return url
            return False
        if url["host4"]:
            return url

        try:
            hostname = url["hostn"].encode("ascii").decode("ascii")
        except UnicodeError:
            try:
                hostname = url["hostn"].encode("idna").decode("ascii")
            except UnicodeError:
                return False

        if hostname[-1] == ".":
            hostname = hostname[:-1]
        if len(hostname) > 253:
            return False

        labels = hostname.split(".")
        for label in labels:
            if not 0 < len(label) < 64:
                return False
            if "-" in (label[0], label[-1]):
                return False
        if self.fqdn:
            if len(labels) == 1 or not self.TLD_REGEX.match(labels[-1]):
                return False

        url["hostn_enc"] = hostname

        return url

    def validate_(self, value, context=None):
        url = self.valid_url(value)
        if not url:
            raise StopValidationError(self.messages["invalid_url"])
        if self.verify_exists:
            url_string = urlquote(
                urlunsplit(
                    (
                        url["scheme"],
                        (url["host6"] or url["host4"] or url["hostn_enc"])
                        + ":"
                        + (url["port"] or ""),
                        url["path"],
                        url["query"],
                        url["frag"],
                    )
                ).encode("utf-8"),
                safe=VALID_CHAR_STRING,
            )
            try:
                urlopen(url_string)
            except URLError as exc:
                raise StopValidationError(self.messages["not_found"]) from exc


class EmailType(StringType):

    """A field that validates input as an E-Mail-Address."""

    MESSAGES = {"email": _("Not a well-formed email address.")}

    EMAIL_REGEX = re.compile(
        r"""^(
        ( ( [%(atext)s]+ (\.[%(atext)s]+)* ) | ("( [%(qtext)s\s] | \\[%(vchar)s\s] )*") )
        @((?!-)[A-Z0-9-]{1,63}(?<!-)\.)+[A-Z]{2,63})$"""
        % {
            "atext": "-A-Z0-9!#$%&'*+/=?^_`{|}~",
            "qtext": "\x21\x23-\x5B\\\x5D-\x7E",
            "vchar": "\x21-\x7E",
        },
        re.I + re.X,
    )

    def _mock(self, context=None):
        return fill_template("%s@example.com", self.min_length, self.max_length)

    def validate_email(self, value, context=None):
        if not EmailType.EMAIL_REGEX.match(value):
            raise StopValidationError(self.messages["email"])



from schematics.models import Model
from schematics.types import IntType, StringType, URLType
from schematics.types.compound import ListType, ModelType


class TagSerializer(Model):
    id = IntType(required=False)
    title = StringType()


class LinkSerializerMixin(Model):
    """Base Mixin for CRUD operation"""

    title = StringType(max_length=255, min_length=5, required=True)
    url = URLType(required=True)


class LinkCreateSerializer(LinkSerializerMixin):
    """Serializer used during POST"""

    tags = ListType(field=StringType, max_size=3, required=False)


class LinkUpdateSerializer(LinkSerializerMixin):
    """Serializer used during PATCH"""


class LinkReadSerializer(LinkSerializerMixin):
    """Serializer used during GET and after successful link creation"""

    id = IntType(required=False)
    tags = ListType(ModelType(TagSerializer))



from django.conf.urls import include, patterns, url
from django.contrib import admin

admin.autodiscover()

from links.views import LinkDetailView, LinkListView

urlpatterns = patterns(
    "",
    # Examples:
    # url(r'^$', 'linkify.views.home', name='home'),
    # url(r'^blog/', include('blog.urls')),
    url(r"^admin/", include(admin.site.urls)),
    url(r"^links/(?P<pk>\d+)/$", LinkDetailView.as_view(), name="detail-link"),
    url(r"^links/$", LinkListView.as_view(), name="list-link"),
)



import datetime
import operator

import pytest

from schematics.exceptions import (
    ConversionError,
    DataError,
    StopValidationError,
    ValidationError,
)
from schematics.models import Model
from schematics.types import BooleanType, DateTimeType, IntType, StringType
from schematics.types.compound import DictType, ListType, ModelType
from schematics.types.serializable import serializable
from schematics.validate import prepare_validator

future_error_msg = "Future dates are not valid"


def test_choices_validates():
    class Document(Model):
        language = StringType(choices=["en", "de"])

    doc = Document({"language": "de"})
    doc.validate()


def test_validation_fails():
    class Document(Model):
        language = StringType(choices=["en", "de"])

    doc = Document({"language": "fr"})
    with pytest.raises(DataError):
        doc.validate()


def test_choices_validates_with_embedded():
    class Document(Model):
        language = StringType(choices=["en", "de"])
        other = ListType(StringType())

    doc = Document({"language": "de", "other": ["somevalue", "other"]})

    doc.validate()


def test_validation_failes_with_embedded():
    class Document(Model):
        language = StringType(choices=["en", "de"])
        other = ListType(StringType())

    doc = Document({"language": "fr", "other": ["somevalue", "other"]})

    with pytest.raises(DataError):
        doc.validate()


def test_validation_none_fails():
    class Player(Model):
        first_name = StringType(required=True)

    with pytest.raises(DataError) as exception:
        Player({"first_name": None}).validate()

    messages = exception.value.errors
    assert len(messages) == 1  # Only one failure
    assert messages["first_name"][0] == "This field is required."


def test_custom_validators():
    now = datetime.datetime(2012, 1, 1, 0, 0)

    def is_not_future(dt, context=None):
        if dt > now:
            raise ValidationError(future_error_msg)

    def without_context(dt):
        pass

    class TestDoc(Model):
        publish = DateTimeType(validators=[is_not_future, without_context])
        author = StringType(required=True)
        title = StringType(required=False)

    with pytest.raises(DataError) as exception:
        TestDoc(
            {
                "publish": datetime.datetime(2012, 2, 1, 0, 0),
                "author": "Hemingway",
                "title": "Old Man",
            }
        ).validate()

    messages = exception.value.errors
    assert future_error_msg in messages["publish"]


def test_messages_subclassing():
    class MyStringType(StringType):
        MESSAGES = {"required": "Never forget"}

    class TestDoc(Model):
        title = MyStringType(required=True)

    with pytest.raises(DataError) as exception:
        TestDoc({"title": None}).validate()

    messages = exception.value.errors
    assert "Never forget" in messages["title"]


def test_messages_instance_level():
    class TestDoc(Model):
        title = StringType(required=True, messages={"required": "Never forget"})

    with pytest.raises(DataError) as exception:
        TestDoc({"title": None}).validate()

    messages = exception.value.errors
    assert "Never forget" in messages["title"]


def test_model_validators():
    now = datetime.datetime(2012, 1, 1, 0, 0)
    future = now + datetime.timedelta(1)

    assert future > now

    class TestDoc(Model):
        can_future = BooleanType()
        publish = DateTimeType()
        foo = StringType()

        def validate_publish(self, data, dt, context):
            if dt > datetime.datetime(2012, 1, 1, 0, 0) and not data["can_future"]:
                raise ValidationError(future_error_msg)

        def validate_foo(self, data, dt):  # without context param
            pass

    TestDoc({"publish": now}).validate()

    with pytest.raises(DataError):
        TestDoc({"publish": future}).validate()


def test_validator_wrapper():
    def f(x, y):
        pass

    assert prepare_validator(f, 2) is f
    assert prepare_validator(f, 3) is not f


def test_nested_model_validators():
    class SubModel(Model):
        should_raise = BooleanType()

        def validate_should_raise(self, data, value, context):
            if value:
                raise ValidationError("message")

    class MainModel(Model):
        modelfield = ModelType(SubModel)

    with pytest.raises(DataError):
        MainModel({"modelfield": {"should_raise": True}}).validate()

    MainModel({"modelfield": {"should_raise": False}}).validate()


def test_multi_key_validation():
    now = datetime.datetime(2012, 1, 1, 0, 0)
    future = now + datetime.timedelta(1)

    input = str(future.isoformat())

    class GoodModel(Model):
        should_raise = BooleanType(default=True)
        publish = DateTimeType()

        def validate_publish(self, data, dt, context):
            if data["should_raise"] is True:
                raise ValidationError("")
            return dt

    with pytest.raises(DataError):
        GoodModel({"publish": input}).validate()

    GoodModel({"publish": input, "should_raise": False}).validate()

    with pytest.raises(DataError):
        GoodModel({"publish": input, "should_raise": True}).validate()


def test_multi_key_validation_part_two():
    class Signup(Model):
        name = StringType()
        call_me = BooleanType(default=False)

        def validate_call_me(self, data, value, context):
            if data["name"] == "Brad" and value is True:
                raise ValidationError(
                    "I'm sorry I never call people who's name is Brad"
                )
            return value

    Signup({"name": "Brad"}).validate()
    Signup({"name": "Brad", "call_me": False}).validate()

    with pytest.raises(DataError):
        Signup({"name": "Brad", "call_me": True}).validate()


def test_multi_key_validation_fields_order():
    class Signup(Model):
        name = StringType()
        call_me = BooleanType(default=False)

        def validate_name(self, data, value, context):
            if data["name"] == "Brad":
                value = "Joe"
                data["name"] = value
                return value
            return value

        def validate_call_me(self, data, value, context):
            if data["name"] == "Joe":
                raise ValidationError("Don't try to decept me! You're Joe!")
            return value

    Signup({"name": "Tom"}).validate()

    with pytest.raises(DataError):
        Signup({"name": "Brad"}).validate()


def test_validate_discovers_classmethod():
    class Foo(Model):
        foo = StringType()

        @classmethod
        def validate_foo(cls, data, value, context):
            raise ValidationError("I'm a classmethod that should be discovered.")

    with pytest.raises(DataError):
        Foo({"foo": "Bar"}).validate()


def test_basic_error():
    class School(Model):
        name = StringType(required=True)

    school = School()

    with pytest.raises(DataError) as exception:
        school.validate()

    errors = exception.value.errors

    assert "name" in errors
    assert errors["name"] == ["This field is required."]


def test_deep_errors():
    class Person(Model):
        name = StringType(required=True)

    class School(Model):
        name = StringType(required=True)
        headmaster = ModelType(Person, required=True)

    school = School()
    school.name = "Hogwarts"
    school.headmaster = {}

    with pytest.raises(DataError) as exception:
        school.validate()

    errors = exception.value.errors

    assert "headmaster" in errors
    assert "name" in errors["headmaster"]
    assert errors["headmaster"]["name"] == ["This field is required."]


@pytest.mark.parametrize("idx1", (0, 1))
@pytest.mark.parametrize("idx2", (0, 1))
def test_deep_errors_with_lists(idx1, idx2):
    class Person(Model):
        name = StringType(required=True)

    class Course(Model):
        id = StringType(required=True, validators=[])
        attending = ListType(ModelType(Person))

    class School(Model):
        courses = ListType(ModelType(Course))

    valid_data = {
        "courses": [
            {"id": "ENG103", "attending": [{"name": "Danny"}, {"name": "Sandy"}]},
            {"id": "ENG203", "attending": [{"name": "Danny"}, {"name": "Sandy"}]},
        ]
    }

    school = School(valid_data)
    school.validate()

    invalid_data = school.serialize()
    invalid_data["courses"][idx1]["attending"][idx2]["name"] = None

    school = School(invalid_data)
    with pytest.raises(DataError) as exception:
        school.validate()

    messages = exception.value.errors

    assert messages == {
        "courses": {
            idx1: {
                "attending": {
                    idx2: {
                        "name": ["This field is required."],
                    },
                },
            }
        }
    }


def test_merge_serializable_errors():
    now = datetime.datetime(2012, 1, 1, 0, 0)

    class Person(Model):
        name = StringType(required=True)
        birthday = DateTimeType()

        @serializable
        def age(self):
            return (now - self.birthday).days / 365

        @age.setter
        def age(self, value):
            if value < 0:
                raise ValidationError("Negative age is not valid")
            self.birthday = now - datetime.timedelta(value * 365)

    person = Person({"age": -1})
    with pytest.raises(DataError) as exception:
        person.validate()

    messages = exception.value.errors

    assert "age" in messages
    assert "name" in messages


def test_deep_errors_with_dicts():
    class Person(Model):
        name = StringType(required=True)

    class Course(Model):
        id = StringType(required=True, validators=[])
        attending = ListType(ModelType(Person))

    class School(Model):
        courses = DictType(ModelType(Course))

    valid_data = {
        "courses": {
            "ENG103": {
                "id": "ENG103",
                "attending": [
                    {"name": "Danny"},
                    {"name": "Sandy"},
                ],
            },
            "ENG203": {
                "id": "ENG203",
                "attending": [
                    {"name": "Danny"},
                    {"name": "Sandy"},
                ],
            },
        }
    }

    school = School(valid_data)
    school.validate()

    invalid_data = school.serialize()
    invalid_data["courses"]["ENG103"]["attending"][0]["name"] = None

    school = School(invalid_data)
    with pytest.raises(DataError) as exception:
        school.validate()

    messages = exception.value.errors

    assert messages == {
        "courses": {"ENG103": {"attending": {0: {"name": ["This field is required."]}}}}
    }


def test_custom_validator_raising_dicterror():
    class Foo(Model):
        bar = DictType(IntType(required=True), required=True)
        eggs = IntType(required=True, min_value=10)

        @classmethod
        def validate_bar(cls, data, value):
            errors = {}

            if "a" not in value:
                errors["a"] = ValidationError("Key a must be set.")
            if "b" not in value:
                errors["b"] = ValidationError("Key b must be set.")

            if errors:
                raise DataError(errors)

    valid_data = {
        "bar": {
            "a": 1,
            "b": 1,
        },
        "eggs": 10,
    }
    foo = Foo(valid_data)
    foo.validate()

    invalid_data = foo.serialize()
    del invalid_data["bar"]["a"]
    invalid_data["eggs"] = 1

    foo = Foo(invalid_data)
    with pytest.raises(DataError) as exception:
        foo.validate()

    messages = exception.value.errors

    assert messages == {
        "bar": {"a": ["Key a must be set."]},
        "eggs": ["Int value should be greater than or equal to 10."],
    }


def test_type_validator_inheritance():
    class CustomIntType(IntType):
        def validate_(self, value, context=None):
            pass

        def validate_foo(self, value, context=None):
            pass

    class SpecialIntType(CustomIntType):
        def validate_range(self, value, context=None):
            pass

        def validate_bar(self, value, context=None):
            pass

    validators_custom = list(
        map(operator.attrgetter("__name__"), CustomIntType().validators)
    )
    validators_special = list(
        map(operator.attrgetter("__name__"), SpecialIntType().validators)
    )

    assert (
        validators_custom[0:2]
        == validators_special[0:2]
        == ["validate_choices", "validate_range"]
    )  # from BaseType & IntType

    assert (
        set(validators_custom[2:4])
        == set(validators_special[2:4])
        == set(["validate_", "validate_foo"])
    )  # both from CustomIntType; mutual order is random

    assert validators_custom[4:] == []
    assert validators_special[4:] == ["validate_bar"]

    assert SpecialIntType(max_value=1).validate(
        9
    )  # Overridden `validate_range()` does nothing.


def test_model_validator_override():
    class Base(Model):
        def validate_foo(self, data, value, context=None):
            pass

        def validate_bar(self, data, value, context=None):
            pass

    class Child(Base):
        def validate_bar(self, data, value, context=None):
            pass

    assert Child._schema.validators["foo"] is Base._schema.validators["foo"]
    assert Child._schema.validators["bar"] is not Base._schema.validators["bar"]


def test_validate_convert():
    class M(Model):
        field1 = IntType()

    m = M()
    m.field1 = "1"
    m.validate()
    assert m.field1 == 1

    m = M()
    m.field1 = "foo"
    m.validate(convert=False)
    assert m.field1 == "foo"


def test_validate_apply_defaults():
    class M(Model):
        field1 = StringType()
        field2 = StringType(default="foo")

    m = M({"field1": None}, init=False)

    m.validate()
    assert m.to_primitive() == {"field1": None}

    m.validate(apply_defaults=True)
    assert m.to_primitive() == {"field1": None, "field2": "foo"}


def test_clean_validation_messages():
    error = ValidationError("A")
    assert error.errors == ["A"]


def test_clean_validation_messages_list():
    error = ValidationError(["A", "B", "C"])
    assert error.errors, ["A", "B", "C"]


def test_builtin_conversion_exception():
    with pytest.raises(TypeError):
        raise ConversionError("message")


def test_builtin_validation_exception():
    with pytest.raises(ValueError):
        raise ValidationError("message")


def test_lazy_conversion_exception():
    class Foo(Model):
        bar = BooleanType()

    foo = Foo(dict(bar="a"), lazy=True)
    with pytest.raises(DataError):
        foo.validate()


def test_lazy_conversion_and_validation_exception_bundling():
    class Signup(Model):
        name = StringType(max_length=3)
        call_me = BooleanType()

    user = Signup({"name": "Arthur", "call_me": "Dent"}, lazy=True)
    with pytest.raises(DataError) as exception:
        user.validate()
    assert set(("name", "call_me")).issubset(exception.value.errors)


def test_validate_required_on_init():
    class Foo(Model):
        bar = BooleanType(required=True)
        baz = BooleanType()

    foo = Foo({"baz": 0}, partial=True, validate=True)
    foo.bar = 1
    assert foo.serialize() == {"bar": True, "baz": False}



import pytest

from schematics.datastructures import Context
from schematics.exceptions import ConversionError, DataError, ValidationError
from schematics.models import Model
from schematics.types import *


def test_ipv4_type():
    assert IPv4Type().validate("255.255.255.255")
    with pytest.raises(ValidationError):
        IPv4Type().validate("1")
    with pytest.raises(ValidationError):
        IPv4Type().validate("255.256.255.255")
    with pytest.raises(ValidationError):
        IPv4Type().validate("255.255.255.2555")

    mock = IPv4Type(required=True).mock()
    assert IPv4Type().validate(mock)


def test_ipv6_type():
    field = IPv6Type()

    addrs = [
        "fe80::223:6caf:fe76:c12d",
        "2001:14ba:ff:a000:223:6caf:fe76:c12d",
        "::255.255.255.255",
        "::1",
    ]
    for addr in addrs:
        field.validate(addr)

    addrs = [
        "",
        "::255.256.255.255",
        ":255.255.255.255",
        "2001:ff:a000:223:6caf:fe76:c12d",
        "fe80::223:6caff:fe76:c12d",
    ]
    for addr in addrs:
        with pytest.raises(ValidationError):
            field.validate(addr)

    mock = IPv6Type(required=True).mock()
    assert IPv6Type().validate(mock)


def test_ip_type():
    assert IPAddressType().validate("255.255.255.255")
    assert IPAddressType().validate("fe80::223:6caf:fe76:c12d")

    mock = IPAddressType(required=True).mock()
    assert IPAddressType().validate(mock)


def test_mac_type():
    addrs = [
        "00-00-00-00-00-00",
        "03:0F:25:B7:10:1E",
        "030F25B7104E",
        "030F25:B7104E",
        "030F25-B7104E",
        "030F.25B7.104E",
    ]
    for addr in addrs:
        assert MACAddressType().validate(addr)

    addrs = [
        "00-00-00-00-00",
        "00:00-00-00-00-00",
        "00:00-00-00-00-00",
        "030F25B7104",
        "030F25B7104Z",
        "30F25:B7104E",
        "030F2-B7104E",
        "030F:25B7.104E",
    ]
    for addr in addrs:
        with pytest.raises(ValidationError):
            MACAddressType().validate(addr)

    mock = MACAddressType(required=True).mock()
    assert MACAddressType().validate(mock)

    s = MACAddressType().to_primitive(value="00-00-00-00-00-00")
    assert MACAddressType().validate(s)


def test_url_type_with_valid_urls():

    field = URLType()
    urls = [
        "https://x." + "x" * 63 + ".com",
        "https://123456789." + ("x" * 59 + ".") * 4 + "com",  # len = 253
        "https://123456789." + ("x" * 59 + ".") * 4 + "com.",  # len = 253 + '.'
        "https://example.fi",
        "http://foo-bar.example.com",
        "HTTP://example.com:80",
        "http://-user:123:%:456(z)@example.com:80",
        "http://example.com/a/b/../c+d/e;f/~jdoe/@?q(x=1;y=2)&r=0#yo!",
        "http://example.com./a/",
        "http://crème-brûlée.tld/menu/à%20la%20carte/",
        "http://はじめよう.みんな",
        "http://xn--p8j9a0d9c9a.xn--q9jyb4c",
        "http://∫ç√œΩ@example.com/?µ=0.3&∂=0.1",
        "http://user:123@127.0.0.1",
        "http://127.0.0.1:99999/",
        "http://127.0.0.1:99999/qweasd",
        "http://[2001:4802:7901::e60a:1375:0:5]",
        "http://[2001:4802:7901::e60a:1375:0:5]:99999",
    ]
    for url in urls:
        field.validate(url)

    field = URLType(fqdn=False)
    urls = [
        "https://1",
        "https://111.q2w",
        "https://localhost",
    ]
    for url in urls:
        field.validate(url)


def test_url_type_with_invalid_url():

    field = URLType()
    urls = [
        "https://1",
        "https://111.q2w",
        "https://localhost",
        "http:example.com",
        "https://example.f",
        "https://example.fi0",
        "ftp://example.com",
        "https://x." + "x" * 64 + ".com",
        "https://1234567890." + ("x" * 59 + ".") * 4 + "com",  # len = 254
        "http://-foobar.example.com",
        "http://qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqq-.example.com",
        "http://example.com../a/",
        "http://ex..ample.com/a/",
        "http://.example.com/a/",
        "http://exam%70le.com/a/",
        "http://example.com|/a/",
        "http://example.com/a b/",
        "http://foo_bar.example.com",
        "http://xn--abcdäedfg.xn--q9jyb4c",  # ACE prefix + non-ASCII character
        "http://example.com/a/\x7F",  # illegal ASCII character
        "http://127.0.0.1:999999/",
        "http://2001:4802:7901::e60a:1375:0:5",
    ]
    for url in urls:
        with pytest.raises(ValidationError):
            field.validate(url)


def test_url_type_with_unreachable_url():
    with pytest.raises(ValidationError):
        URLType(verify_exists=True).validate("http://127.0.0.1:99999/")


def test_email_type_with_valid_addresses():
    field = EmailType()
    addrs = [
        r'"()\\\<>[]:,;@!\"#$%&*+-/=?^_`{}|~.a"@example.org',
        '"foo bar baz"@example.org',
        "Z@foo.zz",
        "123.qwe.asd@foo.bar.baz",
    ]
    for addr in addrs:
        field.validate(addr)


def test_email_type_with_invalid_addresses():
    field = EmailType()
    addrs = [
        r'"qweasd\"@example.org',
        '"qwe"asd"@example.org',
        "curaçao@example.org",
        "foo@local",
    ]
    for addr in addrs:
        with pytest.raises(ValidationError):
            field.validate(addr)



from collections.abc import Set


class Role(Set):

    """
    A ``Role`` object can be used to filter specific fields against a sequence.

    The ``Role`` contains two things: a set of names and a function.
    The function describes how to filter, taking a field name as input and then
    returning ``True`` or ``False`` to indicate that field should or should not
    be skipped.

    A ``Role`` can be operated on as a ``Set`` object representing the fields
    it has an opinion on.  When Roles are combined with other roles, only the
    filtering behavior of the first role is used.
    """

    def __init__(self, function, fields):
        self.function = function
        self.fields = set(fields)

    def _from_iterable(self, iterable):
        return Role(self.function, iterable)

    def __contains__(self, value):
        return value in self.fields

    def __iter__(self):
        return iter(self.fields)

    def __len__(self):
        return len(self.fields)

    def __eq__(self, other):
        return (
            self.function.__name__ == other.function.__name__
            and self.fields == other.fields
        )

    def __str__(self):
        fields = ", ".join(f"'{f}'" for f in self.fields)
        return f"{self.function.__name__}({fields})"

    def __repr__(self):
        return f"<Role {self}>"

    # edit role fields
    def __add__(self, other):
        fields = self.fields.union(other)
        return self._from_iterable(fields)

    def __sub__(self, other):
        fields = self.fields.difference(other)
        return self._from_iterable(fields)

    # apply role to field
    def __call__(self, name, value):
        return self.function(name, value, self.fields)

    # static filter functions
    @staticmethod
    def wholelist(name, value, seq):
        """
        Accepts a field name, value, and a field list.  This function
        implements acceptance of all fields by never requesting a field be
        skipped, thus returns False for all input.

        :param name:
            The field name to inspect.
        :param value:
            The field's value.
        :param seq:
            The list of fields associated with the ``Role``.
        """
        return False

    @staticmethod
    def whitelist(name, value, seq):
        """
        Implements the behavior of a whitelist by requesting a field be skipped
        whenever its name is not in the list of fields.

        :param name:
            The field name to inspect.
        :param value:
            The field's value.
        :param seq:
            The list of fields associated with the ``Role``.
        """

        if seq is not None and len(seq) > 0:
            return name not in seq
        return True

    @staticmethod
    def blacklist(name, value, seq):
        """
        Implements the behavior of a blacklist by requesting a field be skipped
        whenever its name is found in the list of fields.

        :param name:
            The field name to inspect.
        :param value:
            The field's value.
        :param seq:
            The list of fields associated with the ``Role``.
        """
        if seq is not None and len(seq) > 0:
            return name in seq
        return False



import functools

from ..transforms import convert, to_primitive
from ..validate import validate


def _callback_wrap(data, schema, transform, *args, **kwargs):
    return transform(schema, data, *args, **kwargs)


class Machine:
    """A poor man's state machine."""

    states = ("raw", "converted", "validated", "serialized")
    transitions = (
        {"trigger": "init", "to": "raw"},
        {"trigger": "convert", "from": "raw", "to": "converted"},
        {"trigger": "validate", "from": "converted", "to": "validated"},
        {"trigger": "serialize", "from": "validated", "to": "serialized"},
    )
    callbacks = {
        "convert": functools.partial(_callback_wrap, transform=convert, partial=True),
        "validate": functools.partial(
            _callback_wrap, transform=validate, convert=False, partial=False
        ),
        "serialize": functools.partial(_callback_wrap, transform=to_primitive),
    }

    def __init__(self, data, *args):
        self.state = self._transition(trigger="init")["to"]
        self.data = data
        self.args = args

    def __getattr__(self, name):
        return functools.partial(self.trigger, name)

    def _transition(self, trigger=None, src_state=None, dst_state=None):
        try:
            return next(
                self._transitions(
                    trigger=trigger, src_state=src_state, dst_state=dst_state
                )
            )
        except StopIteration:
            return None

    def _transitions(self, trigger=None, src_state=None, dst_state=None):
        def pred(d, key, var):
            return d.get(key) == var if var is not None else True

        return (
            d
            for d in self.transitions
            if pred(d, "trigger", trigger)
            and pred(d, "from", src_state)
            and pred(d, "to", dst_state)
        )

    def trigger(self, trigger):
        transition = self._transition(trigger=trigger, src_state=self.state)
        if not transition:
            raise AttributeError(trigger)
        callback = self.callbacks.get(trigger)
        self.data = callback(self.data, *self.args) if callback else self.data
        self.state = transition["to"]

    def can(self, state):
        return bool(self._transition(src_state=self.state, dst_state=state))

    def cannot(self, state):
        return not self.can(state)



"""Schema validation."""

import functools
import inspect
from typing import List

from .datastructures import Context
from .exceptions import DataError, FieldError
from .iteration import atoms
from .transforms import import_loop, validation_converter
from .undefined import Undefined

__all__: List[str] = []


def schema_from(obj):
    try:
        return obj._schema
    except AttributeError:
        return obj


def validate(
    schema,
    mutable,
    raw_data=None,
    trusted_data=None,
    partial=False,
    strict=False,
    convert=True,
    context=None,
    **kwargs
):
    """
    Validate some untrusted data using a model. Trusted data can be passed in
    the `trusted_data` parameter.

    :param schema:
        The Schema to use as source for validation.
    :param mutable:
        A mapping or instance that can be changed during validation by Schema
        functions.
    :param raw_data:
        A mapping or instance containing new data to be validated.
    :param partial:
        Allow partial data to validate; useful for PATCH requests.
        Essentially drops the ``required=True`` arguments from field
        definitions. Default: False
    :param strict:
        Complain about unrecognized keys. Default: False
    :param trusted_data:
        A ``dict``-like structure that may contain already validated data.
    :param convert:
        Controls whether to perform import conversion before validating.
        Can be turned off to skip an unnecessary conversion step if all values
        are known to have the right datatypes (e.g., when validating immediately
        after the initial import). Default: True

    :returns: data
        ``dict`` containing the valid raw_data plus ``trusted_data``.
        If errors are found, they are raised as a ValidationError with a list
        of errors attached.
    """
    if raw_data is None:
        raw_data = mutable

    context = context or get_validation_context(
        partial=partial, strict=strict, convert=convert
    )

    errors = {}
    try:
        data = import_loop(
            schema,
            mutable,
            raw_data,
            trusted_data=trusted_data,
            context=context,
            **kwargs
        )
    except DataError as exc:
        errors = dict(exc.errors)
        data = exc.partial_data

    errors.update(_validate_model(schema, mutable, data, context))

    if errors:
        raise DataError(errors, data)

    return data


def _validate_model(schema, mutable, data, context):
    """
    Validate data using model level methods.

    :param schema:
        The Schema to validate ``data`` against.
    :param mutable:
        A mapping or instance that will be passed to the validator containing
        the original data and that can be mutated.
    :param data:
        A dict with data to validate. Invalid items are removed from it.
    :returns:
        Errors of the fields that did not pass validation.
    """
    errors = {}
    invalid_fields = []

    def has_validator(atom):
        return (
            atom.value is not Undefined and atom.name in schema_from(schema).validators
        )

    for field_name, field, value in atoms(schema, data, filter=has_validator):
        try:
            schema_from(schema).validators[field_name](mutable, data, value, context)
        except (FieldError, DataError) as exc:
            serialized_field_name = field.serialized_name or field_name
            errors[serialized_field_name] = exc.errors
            invalid_fields.append(field_name)

    for field_name in invalid_fields:
        data.pop(field_name)

    return errors


def get_validation_context(**options):
    validation_options = {
        "field_converter": validation_converter,
        "partial": False,
        "strict": False,
        "convert": True,
        "validate": True,
        "new": False,
    }
    validation_options.update(options)
    return Context(**validation_options)


def prepare_validator(func, argcount):
    if isinstance(func, classmethod):
        func = func.__get__(object).__func__
    func_args = inspect.getfullargspec(func).args
    if len(func_args) < argcount:

        @functools.wraps(func)
        def newfunc(*args, **kwargs):
            sentinel = object()
            if not kwargs or kwargs.pop("context", sentinel) is sentinel:
                args = args[:-1]
            return func(*args, **kwargs)

        return newfunc
    return func



from copy import copy

import pytest

from schematics.exceptions import *
from schematics.models import Model
from schematics.transforms import convert
from schematics.types import *
from schematics.types.compound import *
from schematics.undefined import Undefined


def missing(d, fieldname, init_to_none):
    if init_to_none:
        return d[fieldname] is None
    else:
        return fieldname not in d


def autofail(value, context):
    if value != [42]:
        raise ValidationError("Error!", info=99)


class M(Model):
    intfield = IntType(max_value=2)
    reqfield = StringType(required=True)
    matrixfield = ListType(ListType(IntType(max_value=2)))
    listfield = ListType(IntType(), max_size=3, validators=[autofail])
    modelfield = ModelType("M")


def get_input_dict(variant):

    inputdict = {
        "intfield": "1",
        "reqfield": "foo",
        "listfield": [],
        "modelfield": {
            "reqfield": "bar",
            "listfield": [1, 2, 3, 4],
            "modelfield": {
                "intfield": "3",
                "matrixfield": [[0, 1, 0, 1], [1, 2, 3, 4], ["1", "0", "1", "0"]],
                "listfield": None,
                "modelfield": {"intfield": "0", "reqfield": "foo", "listfield": None},
            },
        },
    }

    if variant == "noerrors":
        del inputdict["listfield"]
        del inputdict["modelfield"]["modelfield"]["intfield"]
        del inputdict["modelfield"]["modelfield"]["matrixfield"]
        inputdict["modelfield"]["listfield"] = [42]
        inputdict["modelfield"]["modelfield"]["reqfield"] = "xyz"

    return inputdict


def get_input_instance(input_init, variant):

    inputinstance = M(init=input_init)
    inputinstance.intfield = "1"
    inputinstance.reqfield = "foo"
    inputinstance.listfield = []
    inputinstance.modelfield = M(init=input_init)
    inputinstance.modelfield.reqfield = "bar"
    inputinstance.modelfield.listfield = [1, 2, 3, 4]
    inputinstance.modelfield.modelfield = M(init=input_init)
    inputinstance.modelfield.modelfield.intfield = "3"
    inputinstance.modelfield.modelfield.matrixfield = [
        [0, 1, 0, 1],
        [1, 2, 3, 4],
        ["1", "0", "1", "0"],
    ]
    inputinstance.modelfield.modelfield.listfield = None
    inputinstance.modelfield.modelfield.modelfield = M(init=input_init)
    inputinstance.modelfield.modelfield.modelfield.intfield = "0"
    inputinstance.modelfield.modelfield.modelfield.reqfield = "foo"
    inputinstance.modelfield.modelfield.modelfield.listfield = None

    if variant == "noerrors":
        del inputinstance.listfield
        del inputinstance.modelfield.listfield
        del inputinstance.modelfield.modelfield.intfield
        del inputinstance.modelfield.modelfield.matrixfield
        inputinstance.modelfield.listfield = [42]
        inputinstance.modelfield.modelfield.reqfield = "xyz"

    return inputinstance


@pytest.fixture
def input(input_instance, input_init, variant):
    if input_instance:
        return get_input_instance(input_init, variant)
    else:
        return get_input_dict(variant)


@pytest.mark.parametrize("variant", (None,))
@pytest.mark.parametrize(
    "input_instance, input_init, init",
    [
        (False, None, True),
        (False, None, False),
        (True, False, True),
        (True, False, False),
        (True, True, True),
        (True, True, False),
    ],
)
def test_conversion(input, input_instance, input_init, init):

    init_to_none = input_init or init

    orig_input = copy(input)

    m = M(input, init=init)

    assert type(m.intfield) is int
    assert type(m.modelfield.modelfield.intfield) is int
    assert type(m.modelfield.modelfield.matrixfield[2][3]) is int
    assert type(m.listfield) is list
    assert type(m.modelfield) is M
    assert type(m.modelfield.modelfield) is M
    assert type(m.modelfield.modelfield.modelfield) is M
    assert type(m.modelfield.listfield) is list
    assert type(m.modelfield.modelfield.matrixfield) is list
    assert type(m.modelfield.modelfield.matrixfield[2]) is list

    assert m._data["listfield"] == []
    assert missing(m.modelfield._data, "intfield", init_to_none)
    assert m.modelfield.modelfield._data["listfield"] is None
    assert missing(m.modelfield.modelfield._data, "reqfield", init_to_none)

    assert input == orig_input

    if input_instance:
        assert m.modelfield is not input.modelfield
        assert m._data["modelfield"] is not input._data["modelfield"]
        assert m.modelfield.listfield is not input.modelfield.listfield
    else:
        assert m.modelfield.listfield is not input["modelfield"]["listfield"]


@pytest.mark.parametrize("variant", (None,))
@pytest.mark.parametrize(
    "input_instance, input_init, init",
    [
        (False, None, True),
        (False, None, False),
        (True, False, True),
        (True, False, False),
        (True, True, True),
        (True, True, False),
    ],
)
def test_conversion_to_dictl(input, input_instance, input_init, init):

    init_to_none = input_init or init

    orig_input = copy(input)

    m = convert(M, input, init_values=init, partial=True)

    assert type(m) is dict
    assert type(m["intfield"]) is int
    assert type(m["modelfield"]["modelfield"]["intfield"]) is int
    assert type(m["modelfield"]["modelfield"]["matrixfield"][2][3]) is int
    assert type(m["listfield"]) is list
    assert type(m["modelfield"]) is dict
    assert type(m["modelfield"]["modelfield"]) is dict
    assert type(m["modelfield"]["modelfield"]["modelfield"]) is dict
    assert type(m["modelfield"]["listfield"]) is list
    assert type(m["modelfield"]["modelfield"]["matrixfield"]) is list
    assert type(m["modelfield"]["modelfield"]["matrixfield"][2]) is list

    assert m["listfield"] == []
    assert missing(m["modelfield"], "intfield", init_to_none)
    assert m["modelfield"]["modelfield"]["listfield"] is None
    assert missing(m["modelfield"]["modelfield"], "reqfield", init_to_none)

    assert input == orig_input

    if input_instance:
        assert m["modelfield"] is not input["modelfield"]

    assert m["modelfield"]["listfield"] is not input["modelfield"]["listfield"]


@pytest.mark.parametrize("variant", (None, "noerrors"))
@pytest.mark.parametrize("partial", (True, False))
@pytest.mark.parametrize(
    "import_, two_pass, input_instance, input_init, init",
    [
        (True, False, False, None, True),
        (True, False, False, None, False),
        (True, False, True, False, True),
        (True, False, True, False, False),
        (True, False, True, True, True),
        (True, False, True, True, False),
        (True, True, False, None, True),
        (True, True, False, None, False),
        (True, True, True, False, True),
        (True, True, True, False, False),
        (True, True, True, True, True),
        (True, True, True, True, False),
        (False, None, True, False, True),
        (False, None, True, False, False),
        (False, None, True, True, True),
        (False, None, True, True, False),
    ],
)
def test_conversion_with_validation(
    input, import_, two_pass, input_instance, input_init, init, partial, variant
):

    init_to_none = input_init or init

    if variant == "noerrors":

        orig_input = copy(input)

        if input_instance:
            assert input.modelfield is orig_input.modelfield

        if import_:
            if two_pass:
                m = M(input, init=init)
                m.validate(partial=partial)
            else:
                m = M(input, init=init, partial=partial, validate=True)
        else:
            input.validate(init_values=init, partial=partial)
            m = input

        assert input == orig_input

        if input_instance:
            if import_:
                assert m.modelfield is not input.modelfield
                assert m._data["modelfield"] is not input._data["modelfield"]
                assert m.modelfield.listfield is not input.modelfield.listfield
            else:
                assert m.modelfield is input.modelfield
                assert m._data["modelfield"] is input._data["modelfield"]
                assert m.modelfield.listfield is input.modelfield.listfield

        return

    if init_to_none:
        partial_data = {
            "intfield": 1,
            "reqfield": "foo",
            "matrixfield": None,
            "modelfield": {
                "intfield": None,
                "reqfield": "bar",
                "matrixfield": None,
                "modelfield": {
                    "reqfield": None,
                    "listfield": None,
                    "modelfield": M(
                        {"intfield": 0, "reqfield": "foo", "listfield": None}
                    ),
                },
            },
        }
    else:
        partial_data = {
            "intfield": 1,
            "reqfield": "foo",
            "modelfield": {
                "reqfield": "bar",
                "modelfield": {
                    "listfield": None,
                    "modelfield": M(
                        {"intfield": 0, "reqfield": "foo", "listfield": None},
                        init=False,
                    ),
                },
            },
        }

    with pytest.raises(DataError) as excinfo:
        if import_:
            if two_pass:
                m = M(input, init=init)
                m.validate(partial=partial)
            else:
                # variant = none
                # partial = true
                M(input, init=init, partial=partial, validate=True)
        else:
            input.validate(init_values=init, partial=partial)

    errors = excinfo.value.errors

    err_list = errors["listfield"]
    assert type(err_list) is ValidationError
    assert len(err_list) == 1

    err_list = errors["modelfield"]["listfield"]
    assert type(err_list) is ValidationError
    assert len(err_list) == 2

    err_list = errors["modelfield"]["modelfield"]["intfield"]
    assert len(err_list) == 1

    if not partial:
        err_list = errors["modelfield"]["modelfield"]["reqfield"]
        assert len(err_list) == 1
        if init_to_none:
            partial_data["modelfield"]["modelfield"].pop("reqfield")

    err_dict = errors["modelfield"]["modelfield"]["matrixfield"]
    sub_err_dict = err_dict[1]
    assert list((k, type(v)) for k, v in sub_err_dict.items()) == [
        (2, ValidationError),
        (3, ValidationError),
    ]
    assert len(err_dict) == 1
    assert len(errors["modelfield"]["modelfield"]) == 2 + (0 if partial else 1)
    assert len(errors["modelfield"]) == 2
    assert len(errors) == 2

    assert excinfo.value.partial_data == partial_data



import pytest

from schematics.common import *
from schematics.models import Model
from schematics.types import *
from schematics.types.compound import *
from schematics.types.serializable import serializable
from schematics.undefined import Undefined

params = [
    (None, None),
    (None, ALL),
    (ALL, ALL),
    (DEFAULT, DEFAULT),
    (NOT_NONE, NOT_NONE),
    (NONEMPTY, NONEMPTY),
]


@pytest.fixture(params=params)
def models(request):

    m_level, n_level = request.param

    class N(Model):
        subfield1 = StringType()
        subfield2 = StringType()
        subfield3 = StringType()
        subfield4 = StringType()

    class M(Model):
        field0 = StringType(export_level=DROP)
        field1 = StringType()
        field2 = StringType()
        field3 = StringType()
        field4 = StringType()
        field5 = StringType(export_level=NOT_NONE)
        field6 = StringType(export_level=ALL)
        listfield = ListType(StringType())
        modelfield1 = ModelType(N)
        modelfield2 = ModelType(N)

    if m_level:
        M._schema.options.export_level = m_level

    if n_level:
        N._schema.options.export_level = n_level

    return M, N


input = {
    "field0": "foo",
    "field1": "foo",
    "field2": "",
    "field3": None,
    "field5": None,
    "listfield": [],
    "modelfield1": {
        "subfield1": "foo",
        "subfield2": "",
        "subfield3": None,
    },
    "modelfield2": {},
}


def test_export_level(models):

    M, N = models
    m_level = M._schema.options.export_level
    n_level = N._schema.options.export_level

    output = M(input, init=False).to_primitive()

    if m_level == NONEMPTY and n_level == NONEMPTY:
        assert output == {
            "field1": "foo",
            "field2": "",
            "field6": None,
            "modelfield1": {
                "subfield1": "foo",
                "subfield2": "",
            },
        }
    elif m_level == NOT_NONE and n_level == NOT_NONE:
        assert output == {
            "field1": "foo",
            "field2": "",
            "field6": None,
            "listfield": [],
            "modelfield1": {
                "subfield1": "foo",
                "subfield2": "",
            },
            "modelfield2": {},
        }
    elif m_level == DEFAULT and n_level == DEFAULT:
        assert output == {
            "field1": "foo",
            "field2": "",
            "field3": None,
            "field6": None,
            "listfield": [],
            "modelfield1": {
                "subfield1": "foo",
                "subfield2": "",
                "subfield3": None,
            },
            "modelfield2": {},
        }
    elif m_level == ALL and n_level == ALL:
        assert output == {
            "field1": "foo",
            "field2": "",
            "field3": None,
            "field4": None,
            "field6": None,
            "listfield": [],
            "modelfield1": {
                "subfield1": "foo",
                "subfield2": "",
                "subfield3": None,
                "subfield4": None,
            },
            "modelfield2": {
                "subfield1": None,
                "subfield2": None,
                "subfield3": None,
                "subfield4": None,
            },
        }
    elif m_level == DEFAULT and n_level == ALL:
        assert output == {
            "field1": "foo",
            "field2": "",
            "field3": None,
            "field6": None,
            "listfield": [],
            "modelfield1": {
                "subfield1": "foo",
                "subfield2": "",
                "subfield3": None,
                "subfield4": None,
            },
            "modelfield2": {
                "subfield1": None,
                "subfield2": None,
                "subfield3": None,
                "subfield4": None,
            },
        }
    else:
        raise Exception(
            f"Assertions missing for testcase m_level={m_level}, n_level={n_level}"
        )


def test_export_level_override(models):

    M, N = models
    m_level = M._schema.options.export_level
    n_level = N._schema.options.export_level

    m = M(input, init=False)

    assert m.to_primitive(export_level=DROP) == {}

    assert m.to_primitive(export_level=NONEMPTY) == {
        "field0": "foo",
        "field1": "foo",
        "field2": "",
        "modelfield1": {
            "subfield1": "foo",
            "subfield2": "",
        },
    }

    assert m.to_primitive(export_level=NOT_NONE) == {
        "field0": "foo",
        "field1": "foo",
        "field2": "",
        "listfield": [],
        "modelfield1": {
            "subfield1": "foo",
            "subfield2": "",
        },
        "modelfield2": {},
    }

    assert m.to_primitive(export_level=DEFAULT) == {
        "field0": "foo",
        "field1": "foo",
        "field2": "",
        "field3": None,
        "field5": None,
        "listfield": [],
        "modelfield1": {
            "subfield1": "foo",
            "subfield2": "",
            "subfield3": None,
        },
        "modelfield2": {},
    }

    assert m.to_primitive(export_level=ALL) == {
        "field0": "foo",
        "field1": "foo",
        "field2": "",
        "field3": None,
        "field4": None,
        "field5": None,
        "field6": None,
        "listfield": [],
        "modelfield1": {
            "subfield1": "foo",
            "subfield2": "",
            "subfield3": None,
            "subfield4": None,
        },
        "modelfield2": {
            "subfield1": None,
            "subfield2": None,
            "subfield3": None,
            "subfield4": None,
        },
    }


def test_custom_converter():
    class M(Model):
        x = IntType()
        y = IntType()
        z = IntType()

    def converter(field, value, context):
        if field.name == "z":
            return Undefined
        else:
            return field.export(value, PRIMITIVE, context)

    m = M(dict(x=1, y=None, z=3))

    assert m.export(field_converter=converter, export_level=DEFAULT) == {
        "x": 1,
        "y": None,
    }



import pytest

from schematics.contrib.enum_type import EnumType
from schematics.exceptions import ConversionError

try:
    from enum import Enum

    class E(Enum):
        A = 1
        B = "b"

    class F(Enum):
        A = 1
        B = 1

except ImportError:
    Enum = None

pytestmark = pytest.mark.skipif(Enum is None, reason="requires enum")


def test_to_native_by_name():
    field = EnumType(E)
    assert field.to_native("A") == E.A
    assert field.to_native("B") == E.B
    with pytest.raises(ConversionError):
        field.to_native("a")


def test_to_native_by_value():
    field = EnumType(E, use_values=True)
    assert field.to_native(1) == E.A
    assert field.to_native("b") == field.to_native("B")
    with pytest.raises(ConversionError):
        field.to_native(2)


def test_to_native_by_value_duplicate():
    field = EnumType(F, use_values=True)
    assert field.to_native(1) == F.A


def test_passthrough():
    field = EnumType(E, use_values=True)
    assert field.to_native(E.A) == E.A


def test_to_primitive_by_name():
    field = EnumType(E, use_values=False)
    assert field.to_primitive(E.A) == "A"


def test_to_primitive_by_value():
    field = EnumType(E, use_values=True)
    assert field.to_primitive(E.A) == 1



import pytest


def test_translator():
    def translator(string):
        translations = {"String value is too long.": "Tamanho de texto muito grande."}
        return translations.get(string, string)

    from schematics.translator import register_translator

    register_translator(translator)

    from schematics.exceptions import ValidationError
    from schematics.types import StringType

    with pytest.raises(ValidationError) as exc:
        StringType(max_length=1).validate_length("Abc")
    assert exc.value == ["Tamanho de texto muito grande."]



import datetime
import uuid

import pytest

from schematics.common import *
from schematics.models import Model
from schematics.transforms import Converter, to_native, to_primitive
from schematics.types import *
from schematics.types.compound import *
from schematics.types.serializable import serializable


class BaseModel(Model):
    pass


class N(BaseModel):
    floatfield = FloatType()
    uuidfield = UUIDType()


class M(BaseModel):
    intfield = IntType()
    stringfield = StringType()
    dtfield = DateTimeType()
    utcfield = UTCDateTimeType()
    modelfield = ModelType(N)


field = ListType(ModelType(M))  # standalone field

primitives = {
    "intfield": 3,
    "stringfield": "foobar",
    "dtfield": "2015-11-26T09:00:00.000000",
    "utcfield": "2015-11-26T07:00:00.000000Z",
    "modelfield": {
        "floatfield": 1.0,
        "uuidfield": "54020382-291e-4192-b370-4850493ac5bc",
    },
}

natives = {
    "intfield": 3,
    "stringfield": "foobar",
    "dtfield": datetime.datetime(2015, 11, 26, 9),
    "utcfield": datetime.datetime(2015, 11, 26, 7),
    "modelfield": {
        "floatfield": 1.0,
        "uuidfield": uuid.UUID("54020382-291e-4192-b370-4850493ac5bc"),
    },
}


def test_to_native():

    m = M(primitives)
    output = m.to_native()
    assert type(output) is dict
    assert output == natives

    assert to_native(M, natives) == natives


def test_to_primitive():

    m = M(primitives)
    output = m.to_primitive()
    assert type(output) is dict
    assert output == primitives

    assert to_primitive(M, natives) == primitives


def test_standalone_field():

    converted = field.convert([primitives])
    assert converted == [natives]
    assert field.to_native(converted) == [natives]
    assert field.to_primitive(converted) == [primitives]


class Foo:
    def __init__(self, x, y):
        self.x, self.y = x, y

    def __eq__(self, other):
        return type(self) == type(other) and self.x == other.x and self.y == other.y


class FooType(BaseType):
    def to_native(self, value, context):
        if isinstance(value, Foo):
            return value
        return Foo(value["x"], value["y"])

    def to_primitive(self, value, context):
        return dict(x=value.x, y=value.y)


def test_custom_exporter():
    class X(Model):
        id = UUIDType()
        dt = UTCDateTimeType()
        foo = FooType()

    x = X(
        {
            "id": "54020382-291e-4192-b370-4850493ac5bc",
            "dt": "2015-11-26T07:00",
            "foo": {"x": 1, "y": 2},
        }
    )

    assert x.to_native() == {
        "id": uuid.UUID("54020382-291e-4192-b370-4850493ac5bc"),
        "dt": datetime.datetime(2015, 11, 26, 7),
        "foo": Foo(1, 2),
    }

    class MyExportConverter(Converter):

        keep_as_native = (UTCDateTimeType, UUIDType)

        def __call__(self, field, value, context):
            if field.typeclass in self.keep_as_native:
                format = NATIVE
            else:
                format = PRIMITIVE
            return field.export(value, format, context)

        def post(self, model_class, data, context):
            data["n"] = 42
            return data

    exporter = MyExportConverter()

    assert x.export(field_converter=exporter) == {
        "id": uuid.UUID("54020382-291e-4192-b370-4850493ac5bc"),
        "dt": datetime.datetime(2015, 11, 26, 7),
        "foo": {"x": 1, "y": 2},
        "n": 42,
    }


def test_converter_function():
    class X(Model):
        id = UUIDType()
        dt = UTCDateTimeType()
        foo = FooType()

    x = X(
        {
            "id": "54020382-291e-4192-b370-4850493ac5bc",
            "dt": "2015-11-26T07:00",
            "foo": {"x": 1, "y": 2},
        }
    )

    exporter = lambda field, value, context: field.export(value, PRIMITIVE, context)

    assert x.export(field_converter=exporter) == x.to_primitive()


def test_export_order():
    class O(Model):
        class Options:
            export_order = True

        f = StringType()
        a = StringType()
        e = StringType()
        b = StringType()
        d = StringType()
        c = StringType()

    assert list(O().to_primitive().keys()) == ["f", "a", "e", "b", "d", "c"]

    class O(Model):
        class Options:
            export_order = True

        b = StringType()
        f = StringType()
        c = StringType()
        e = StringType()
        d = StringType()
        a = StringType()

    assert list(O().to_primitive().keys()) == ["b", "f", "c", "e", "d", "a"]



import pytest

from schematics.contrib.machine import Machine
from schematics.exceptions import DataError
from schematics.models import Model
from schematics.schema import Field, Schema
from schematics.transforms import convert, to_primitive
from schematics.types import IntType, StringType, calculated, serializable
from schematics.validate import validate


@pytest.fixture
def player_schema():
    def get_full_name(data, *a, **kw):
        if not data:
            return
        return "{first_name} {last_name}".format(**data)

    def set_full_name(data, value, *a, **kw):
        if not value:
            return
        data["first_name"], _, data["last_name"] = value.partition(" ")

    schema = Schema(
        "Player",
        Field("id", IntType()),
        Field("first_name", StringType(required=True)),
        Field("last_name", StringType(required=True)),
        Field(
            "full_name",
            calculated(type=StringType(), fget=get_full_name, fset=set_full_name),
        ),
    )

    return schema


@pytest.fixture
def player_data():
    return {"id": "42", "full_name": "Arthur Dent", "towel": True}


def test_functional_schema_required(player_schema):
    with pytest.raises(DataError):
        validate(player_schema, {}, partial=False)


def test_functional_schema(player_schema, player_data):
    schema = player_schema
    input_data = player_data

    data = input_data  # state = 'RAW'

    expected = {"id": 42, "full_name": "Arthur Dent"}
    data = convert(schema, data, partial=True)
    assert data == expected  # state = 'CONVERTED'

    expected = {
        "id": 42,
        "first_name": "Arthur",
        "last_name": "Dent",
        "full_name": "Arthur Dent",
    }
    data = validate(schema, data, convert=False, partial=False)
    assert data == expected  # state = 'VALIDATED'

    expected = {
        "id": 42,
        "first_name": "Arthur",
        "last_name": "Dent",
        "full_name": "Arthur Dent",
    }
    data = to_primitive(schema, data)
    assert data == expected  # state = 'SERIALIZED'


def test_state_machine_equivalence(player_schema, player_data):
    schema = player_schema
    input_data = player_data

    data = input_data.copy()
    machine = Machine(input_data, schema)
    assert machine.state == "raw"

    data = convert(schema, data, partial=True)
    machine.convert()
    assert machine.state == "converted"
    assert data == machine.data

    data = validate(schema, data, convert=False, partial=False)
    machine.validate()
    assert machine.state == "validated"
    assert data == machine.data

    data = to_primitive(schema, data)
    machine.serialize()
    assert machine.state == "serialized"
    assert data == machine.data


def test_object_model_equivalence():
    # functional
    def get_full_name(data, *a, **kw):
        if not data:
            return
        return "{first_name} {last_name}".format(**data)

    def set_full_name(data, value, *a, **kw):
        if not value:
            return
        data["first_name"], _, data["last_name"] = value.partition(" ")

    schema = Schema(
        "Player",
        Field("id", IntType()),
        Field("first_name", StringType(required=True)),
        Field("last_name", StringType(required=True)),
        Field(
            "full_name",
            calculated(type=StringType(), fget=get_full_name, fset=set_full_name),
        ),
    )

    # object
    class Player(Model):
        id = IntType()
        first_name = StringType(required=True)
        last_name = StringType(required=True)

        @serializable(type=StringType())
        def full_name(self):
            return get_full_name(self)

        @full_name.setter
        def full_name(self, value):
            set_full_name(self, value)

    input_data = {"id": "42", "full_name": "Arthur Dent", "towel": True}

    data = input_data.copy()
    player = Player(input_data, strict=False, validate=False, init=False)

    data = convert(schema, data, partial=True)
    assert data == player._data

    data = validate(schema, data, convert=False, partial=False)
    player.validate()
    assert data == player._data

    data = to_primitive(schema, data)
    player_dict = player.serialize()
    assert data == player_dict



import json

from django.http import Http404, JsonResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from schematics.exceptions import ModelConversionError, ModelValidationError

from .models import Link, Tag
from .serializers import LinkCreateSerializer, LinkReadSerializer, LinkUpdateSerializer


class CSRFExemptMixin(View):
    @method_decorator(csrf_exempt)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)


class LinkListView(CSRFExemptMixin):
    http_method_names = ["post", "get"]

    def post(self, request):
        data = json.loads(request.body.decode())
        try:
            link = LinkCreateSerializer(raw_data=data)
            link.validate()
            kwargs = link.to_native()
            # Pop tags since objects will be created separately
            tags = kwargs.pop("tags", None)

            # Persist the data
            link_obj = Link.objects.create(**kwargs)
            tag_collection = [Tag.objects.get_or_create(title=tag)[0] for tag in tags]
            link_obj.attach_tags(tag_collection)

            # Prepare for response
            return_data = LinkReadSerializer(link_obj.to_dict()).to_native()
            return JsonResponse(data=return_data, status=201)
        except (ModelValidationError, ModelConversionError) as e:
            return JsonResponse(e.messages, status=400)

    def get(self, request):
        # TODO: Add pagination
        links = Link.objects.all()
        items = [LinkReadSerializer(link.to_dict()).to_native() for link in links]
        data = {"items": items, "total": len(links)}

        return JsonResponse(data=data)


class LinkDetailView(CSRFExemptMixin):
    http_method_names = ["delete", "get", "patch"]

    def get_or_404(self, pk):
        try:
            return Link.objects.get(pk=pk)
        except Link.DoesNotExist:
            raise Http404("Link doesn't exist")

    def get(self, request, pk):
        link = self.get_or_404(pk=pk)
        return_data = LinkReadSerializer(link.to_dict()).to_native()
        return JsonResponse(return_data)

    def delete(self, request, pk):
        link = self.get_or_404(pk=pk)
        link.delete()
        return JsonResponse(data={}, status=204)

    def patch(self, request, pk):
        data = json.loads(request.body.decode())
        try:
            link = LinkUpdateSerializer(raw_data=data)
            kwargs = link.to_native()
            # We need to make two db calls any way to return Response
            link_obj = self.get_or_404(pk=pk)
            link_obj.update(**kwargs)
            # Prepare response
            data = LinkReadSerializer(link_obj.to_dict()).to_native()
            return JsonResponse(data=data, status=202)
        except ModelConversionError as e:
            # this is raised when fields are missing
            msg = "PATCH not allowed."
            data = {field: msg for field, value in e.messages.items()}
            return JsonResponse(data=data, status=400)
        except ModelValidationError as e:
            return JsonResponse(e.messages, status=400)



"""
Core loop over the data structures according to a defined schema.
"""

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
)

from .undefined import Undefined

if TYPE_CHECKING:
    from schematics.schema import Schema


class Atom(NamedTuple):
    name: Optional[str] = None
    field: Optional[str] = None
    value: Any = None


def schema_from(obj):
    try:
        return obj._schema
    except AttributeError:
        return obj


def atoms(
    schema: "Schema",
    mapping: Mapping,
    keys: Tuple[str, ...] = tuple(Atom._fields),
    filter: Callable[[Atom], bool] = None,
) -> Iterable[Atom]:
    """
    Iterator for the atomic components of a model definition and relevant
    data that creates a 3-tuple of the field's name, its type instance and
    its value.

    :type schema: schematics.schema.Schema
    :param schema:
        The Schema definition.
    :type mapping: Mapping
    :param mapping:
        The structure where fields from schema are mapped to values. The only
        expectation for this structure is that it implements a ``Mapping``
        interface.
    :type keys: Tuple[str, str, str]
    :param keys:
        Tuple specifying the output of the iterator. Valid keys are:
            `name`: the field name
            `field`: the field descriptor object
            `value`: the current value set on the field
        Specifying invalid keys will raise an exception.
    :type filter: Optional[Callable[[Atom], bool]]
    :param filter:
        Function to filter out atoms from the iteration.

    :rtype: Iterable[Atom]
    """
    if not set(keys).issubset(Atom._fields):
        raise TypeError("invalid key specified")

    has_name = "name" in keys
    has_field = "field" in keys
    has_value = (mapping is not None) and ("value" in keys)

    for field_name, field in schema_from(schema).fields.items():
        value = Undefined

        if has_value:
            try:
                value = mapping[field_name]
            except Exception:
                value = Undefined

        atom_tuple = Atom(
            name=field_name if has_name else None,
            field=field if has_field else None,
            value=value,
        )
        if filter is None:
            yield atom_tuple
        elif filter(atom_tuple):
            yield atom_tuple


class atom_filter:
    """Group for the default filter functions."""

    @staticmethod
    def has_setter(atom):
        return getattr(atom.field, "fset", None) is not None

    @staticmethod
    def not_setter(atom):
        return not atom_filter.has_setter(atom)



import itertools
from collections.abc import Iterable, Mapping, Sequence
from typing import Type, TypeVar

from ..common import DROP, NONEMPTY, NOT_NONE
from ..exceptions import BaseError, CompoundError, ConversionError, ValidationError
from ..transforms import (
    convert,
    export_loop,
    get_export_context,
    get_import_context,
    to_native_converter,
    to_primitive_converter,
)
from ..translator import _
from ..util import get_all_subclasses, import_string
from .base import BaseType, get_value_in

T = TypeVar("T")

__all__ = [
    "CompoundType",
    "MultiType",
    "ModelType",
    "ListType",
    "DictType",
    "PolyModelType",
]


class CompoundType(BaseType):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.is_compound = True
        try:
            self.field.parent_field = self
        except AttributeError:
            pass

    def _setup(self, field_name, owner_model):
        # Recursively set up inner fields.
        try:
            field = self.field
        except AttributeError:
            pass
        else:
            field._setup(None, owner_model)
        super()._setup(field_name, owner_model)

    def convert(self, value, context=None):
        context = context or get_import_context()
        return self._convert(value, context)

    def _convert(self, value, context):
        raise NotImplementedError

    def export(self, value, format, context=None):
        context = context or get_export_context()
        return self._export(value, format, context)

    def _export(self, value, format, context):
        raise NotImplementedError

    def to_native(self, value, context=None):
        context = context or get_export_context(to_native_converter)
        return to_native_converter(self, value, context)

    def to_primitive(self, value, context=None):
        context = context or get_export_context(to_primitive_converter)
        return to_primitive_converter(self, value, context)

    def _init_field(self, field, options):
        """
        Instantiate the inner field that represents each element within this compound type.
        In case the inner field is itself a compound type, its inner field can be provided
        as the ``nested_field`` keyword argument.
        """
        if not isinstance(field, BaseType):
            nested_field = options.pop("nested_field", None) or options.pop(
                "compound_field", None
            )
            if nested_field:
                field = field(field=nested_field, **options)
            else:
                field = field(**options)
        return field


MultiType = CompoundType


class ModelType(CompoundType):
    """A field that can hold an instance of the specified model."""

    primitive_type = dict

    @property
    def native_type(self):
        return self.model_class

    @property
    def fields(self):
        return self.model_class.fields

    @property
    def model_class(self):
        if self._model_class:
            return self._model_class

        model_class = import_string(self.model_name)
        self._model_class = model_class
        return model_class

    def __init__(self, model_spec: Type[T], **kwargs):

        from ..models import ModelMeta

        if isinstance(model_spec, ModelMeta):
            self._model_class = model_spec
            self.model_name = self.model_class.__name__
        elif isinstance(model_spec, str):
            self._model_class = None
            self.model_name = model_spec
        else:
            raise TypeError(
                "ModelType: Expected a model, got an argument "
                "of the type '{}'.".format(model_spec.__class__.__name__)
            )

        super().__init__(**kwargs)

    def _repr_info(self):
        return self.model_class.__name__

    def _mock(self, context=None):
        return self.model_class.get_mock_object(context)

    def _setup(self, field_name, owner_model):
        # Resolve possible name-based model reference.
        if not self._model_class:
            if self.model_name == owner_model.__name__:
                self._model_class = owner_model
            else:
                pass  # Intentionally left blank, it will be setup later.
        super()._setup(field_name, owner_model)

    def pre_setattr(self, value):
        from ..models import Model

        if value is not None and not isinstance(value, Model):
            if not isinstance(value, dict):
                raise ConversionError(_("Model conversion requires a model or dict"))
            value = self.model_class(value)
        return value

    def _convert(self, value, context):
        field_model_class = self.model_class
        if isinstance(value, field_model_class):
            model_class = type(value)
        elif isinstance(value, dict):
            model_class = field_model_class
        else:
            raise ConversionError(
                _("Input must be a mapping or '%s' instance")
                % field_model_class.__name__
            )
        if context.convert and context.oo:
            return model_class(value, context=context)
        return convert(model_class._schema, value, context=context)

    def _export(self, value, format, context):

        from ..models import Model

        if isinstance(value, Model):
            model_class = type(value)
        else:
            model_class = self.model_class
        return export_loop(model_class, value, context=context)


class ListType(CompoundType):
    """A field for storing a list of items, all of which must conform to the type
    specified by the ``field`` parameter.

    Use it like this::

        ...
        categories = ListType(StringType)
    """

    primitive_type = list
    native_type = list

    def __init__(self, field: T, min_size=None, max_size=None, **kwargs):
        """Create a list of objects of type `field`."""

        self.field = self._init_field(field, kwargs)
        self.min_size = min_size
        self.max_size = max_size

        validators = [self.check_length] + kwargs.pop("validators", [])

        super().__init__(validators=validators, **kwargs)

    @property
    def model_class(self):
        return self.field.model_class

    def _repr_info(self):
        return self.field.__class__.__name__

    def _mock(self, context=None):
        random_length = get_value_in(self.min_size, self.max_size)

        return [self.field._mock(context) for dummy in range(random_length)]

    def _coerce(self, value):
        if isinstance(value, list):
            return value
        if isinstance(value, (str, Mapping)):  # unacceptable iterables
            pass
        elif isinstance(value, Sequence):
            return value
        elif isinstance(value, Iterable):
            return value
        raise ConversionError(_("Could not interpret the value as a list"))

    def _convert(self, value, context):
        value = self._coerce(value)
        data = []
        errors = {}
        for index, item in enumerate(value):
            try:
                data.append(context.field_converter(self.field, item, context))
            except BaseError as exc:
                errors[index] = exc
        if errors:
            raise CompoundError(errors)
        return data

    def check_length(self, value, context):
        list_length = len(value) if value else 0

        if self.min_size is not None and list_length < self.min_size:
            message = (
                {
                    True: _("Please provide at least %d item."),
                    False: _("Please provide at least %d items."),
                }[self.min_size == 1]
            ) % self.min_size
            raise ValidationError(message)

        if self.max_size is not None and list_length > self.max_size:
            message = (
                {
                    True: _("Please provide no more than %d item."),
                    False: _("Please provide no more than %d items."),
                }[self.max_size == 1]
            ) % self.max_size
            raise ValidationError(message)

    def _export(self, list_instance, format, context):
        """Loops over each item in the model and applies either the field
        transform or the multitype transform.  Essentially functions the same
        as `transforms.export_loop`.
        """
        data = []
        _export_level = self.field.get_export_level(context)
        if _export_level == DROP:
            return data
        for value in list_instance:
            shaped = self.field.export(value, format, context)
            if shaped is None:
                if _export_level <= NOT_NONE:
                    continue
            elif self.field.is_compound and len(shaped) == 0:
                if _export_level <= NONEMPTY:
                    continue
            data.append(shaped)
        return data


class DictType(CompoundType):
    """A field for storing a mapping of items, the values of which must conform to the type
    specified by the ``field`` parameter.

    Use it like this::

        ...
        categories = DictType(StringType)

    """

    primitive_type = dict
    native_type = dict

    def __init__(self, field, coerce_key=None, **kwargs):
        """Create a dict with str keys and type `field` values."""

        self.field = self._init_field(field, kwargs)
        self.coerce_key = coerce_key or str
        super().__init__(**kwargs)

    @property
    def model_class(self):
        return self.field.model_class

    def _repr_info(self):
        return self.field.__class__.__name__

    def _convert(self, value, context, safe=False):
        if not isinstance(value, Mapping):
            raise ConversionError(_("Only mappings may be used in a DictType"))

        data = {}
        errors = {}
        for k, v in value.items():
            try:
                data[self.coerce_key(k)] = context.field_converter(
                    self.field, v, context
                )
            except BaseError as exc:
                errors[k] = exc
        if errors:
            raise CompoundError(errors)
        return data

    def _export(self, dict_instance, format, context):
        """Loops over each item in the model and applies either the field
        transform or the multitype transform.  Essentially functions the same
        as `transforms.export_loop`.
        """
        data = {}
        _export_level = self.field.get_export_level(context)
        if _export_level == DROP:
            return data
        for key, value in dict_instance.items():
            shaped = self.field.export(value, format, context)
            if shaped is None:
                if _export_level <= NOT_NONE:
                    continue
            elif self.field.is_compound and len(shaped) == 0:
                if _export_level <= NONEMPTY:
                    continue
            data[key] = shaped
        return data


class PolyModelType(CompoundType):
    """A field that accepts an instance of any of the specified models."""

    primitive_type = dict
    native_type = None  # cannot be determined from a PolyModelType instance

    def __init__(self, model_spec, **kwargs):

        from ..models import ModelMeta

        if isinstance(model_spec, (ModelMeta, str)):
            self.model_classes = (model_spec,)
            allow_subclasses = True
        elif isinstance(model_spec, Iterable):
            self.model_classes = tuple(model_spec)
            allow_subclasses = False
        else:
            raise Exception(
                "The first argument to PolyModelType.__init__() "
                "must be a model or an iterable."
            )

        self.claim_function = kwargs.pop("claim_function", None)
        self.allow_subclasses = kwargs.pop("allow_subclasses", allow_subclasses)

        CompoundType.__init__(self, **kwargs)

    def _setup(self, field_name, owner_model):
        # Resolve possible name-based model references.
        resolved_classes = []
        for m in self.model_classes:
            if isinstance(m, str):
                if m == owner_model.__name__:
                    resolved_classes.append(owner_model)
                else:
                    raise Exception(
                        "PolyModelType: Unable to resolve model '{}'.".format(m)
                    )
            else:
                resolved_classes.append(m)
        self.model_classes = tuple(resolved_classes)
        super()._setup(field_name, owner_model)

    def is_allowed_model(self, model_instance):
        if self.allow_subclasses:
            if isinstance(model_instance, self.model_classes):
                return True
        else:
            if model_instance.__class__ in self.model_classes:
                return True
        return False

    def _convert(self, value, context):

        if value is None:
            return None

        if not context.validate:
            if self.is_allowed_model(value):
                return value
            if not isinstance(value, dict):
                if len(self.model_classes) > 1:
                    instanceof_msg = "one of: {}".format(
                        ", ".join(cls.__name__ for cls in self.model_classes)
                    )
                else:
                    instanceof_msg = self.model_classes[0].__name__
                raise ConversionError(
                    _(
                        "Please use a mapping for this field or " "an instance of {}"
                    ).format(instanceof_msg)
                )

        model_class = self.find_model(value)
        return model_class(value, context=context)

    def find_model(self, data):
        """Finds the intended type by consulting potential classes or `claim_function`."""

        if self.claim_function:
            kls = self.claim_function(self, data)
            if not kls:
                raise Exception("Input for polymorphic field did not match any model")
            return kls

        fallback = None
        matching_classes = []
        for kls in self._get_candidates():
            try:
                # If a model defines a _claim_polymorphic method, use
                # it to see if the model matches the data.
                kls_claim = kls._claim_polymorphic
            except AttributeError:
                # The first model that doesn't define the hook can be
                # used as a default if there's no match.
                if not fallback:
                    fallback = kls
            else:
                if kls_claim(data):
                    matching_classes.append(kls)

        if not matching_classes and fallback:
            return fallback

        if len(matching_classes) != 1:
            raise Exception("Got ambiguous input for polymorphic field")

        return matching_classes[0]

    def _export(self, model_instance, format, context):

        model_class = model_instance.__class__
        if not self.is_allowed_model(model_instance):
            raise Exception(
                "Cannot export: {} is not an allowed type".format(model_class)
            )

        return model_instance.export(context=context)

    def _get_candidates(self):
        candidates = self.model_classes

        if self.allow_subclasses:
            candidates = itertools.chain.from_iterable(
                ([m] + get_all_subclasses(m) for m in candidates)
            )

        return candidates



import pytest

from schematics.exceptions import *
from schematics.models import Model
from schematics.schema import SchemaOptions
from schematics.transforms import blacklist, whitelist
from schematics.types import IntType, ListType, ModelType, StringType


def test_dict_methods_in_model():
    """
    a regression test to ensure that an issue where attributes on
    dictionaries are not being misintrepreted as actual schematics
    fields.
    """

    class M(Model):
        items, values, get, keys = IntType(), IntType(), IntType(), IntType()

    m = M({"items": 1, "values": 1, "get": 1, "keys": 1})
    m.validate()


def test_dict_methods_in_model_atoms():
    """
    atoms should return the raw values, and not call any overriden methods.
    """

    class M(Model):
        get = IntType()

    m = M({"get": 1})
    atom = list(m.atoms())[0]
    assert atom.name == "get"
    assert atom.value == 1


def test_nested_model_override_mapping_methods():
    """
    overriding mapping methods on child models should not cause issues
    with validation on the parent.
    """

    class Nested(Model):
        items, values, get, keys = IntType(), IntType(), IntType(), IntType()

    class Root(Model):
        keys = ModelType(Nested)

    root = Root({"keys": {"items": 1, "values": 1, "get": 1, "keys": 1}})
    root.validate()
    for key in ["items", "values", "get", "keys"]:
        assert getattr(root.keys, key) == 1


def test_init_with_dict():
    class M(Model):
        a, b, c, d = IntType(), IntType(), IntType(), IntType(default=0)

    m = M({"a": 1, "b": None})
    assert m._data == {"a": 1, "b": None, "c": None, "d": 0}
    assert m.a == 1
    assert m.b == None
    assert m.c == None
    assert m.d == 0

    m = M({"a": 1, "b": None}, apply_defaults=False)
    assert m._data == {"a": 1, "b": None, "c": None, "d": None}

    m = M({"a": 1, "b": None}, init_values=False)
    assert m._data == {"a": 1, "b": None, "d": 0}

    m = M({"a": 1, "b": None}, init=False)
    assert m._data == {"a": 1, "b": None}

    m = M({"a": 1, "b": None}, init=False, apply_defaults=True)
    assert m._data == {"a": 1, "b": None, "d": 0}

    m = M({"a": 1, "b": None}, init=False, init_values=True)
    assert m._data == {"a": 1, "b": None, "c": None, "d": None}

    m = M({"a": 1, "b": None}, init=False, apply_defaults=True, init_values=True)
    assert m._data == {"a": 1, "b": None, "c": None, "d": 0}


def test_defaults():
    class M(Model):
        d0 = IntType(default=0)
        dN = ListType(IntType, default=None)

    m = M()
    assert m._data == {"d0": 0, "dN": None}
    m.validate()
    assert m._data == {"d0": 0, "dN": None}

    m = M(apply_defaults=False)
    assert m._data == {"d0": None, "dN": None}
    m.validate()
    assert m._data == {"d0": None, "dN": None}

    m = M(init_values=False)
    assert m._data == {"d0": 0, "dN": None}
    m.validate()
    assert m._data == {"d0": 0, "dN": None}

    m = M(init=False)
    assert m._data == {}
    m.validate()
    assert m._data == {}

    m = M(init=False, apply_defaults=True)
    assert m._data == {"d0": 0, "dN": None}
    m.validate()
    assert m._data == {"d0": 0, "dN": None}


def test_invalid_model_fail_validation():
    class Player(Model):
        name = StringType(required=True)

    p = Player()
    assert p.name is None

    with pytest.raises(DataError):
        p.validate()


def test_invalid_models_validate_partially():
    class User(Model):
        name = StringType(required=True)

    u = User()
    u.validate(partial=True)


def test_model_with_rogue_field_throws_exception():
    class User(Model):
        name = StringType()

    with pytest.raises(DataError):
        User({"foo": "bar"})


def test_equality():
    class Player(Model):
        id = IntType()

    p1 = Player({"id": 4})
    p2 = Player({"id": 4})

    assert p1 == p2

    p3 = Player({"id": 5})

    assert p1 == p2
    assert p1 != p3


def test_dict_interface():
    class Player(Model):
        name = StringType()

    p = Player()
    p.name = "Jóhann"

    assert "name" in p
    assert p["name"] == "Jóhann"
    assert "fake_key" not in p


def test_init_model_from_another_model():
    class User(Model):
        name = StringType(required=True)
        bio = StringType(required=True)

    u = User(dict(name="A", bio="Asshole"))

    u2 = User(u)
    assert u == u2


def test_raises_validation_error_on_non_partial_validate():
    class User(Model):
        name = StringType(required=True)
        bio = StringType(required=True)

    u = User(dict(name="Joe"))

    with pytest.raises(DataError) as exception:
        u.validate()
    assert exception.value.errors, {"bio": ["This field is required."]}


def test_model_inheritance():
    class Parent(Model):
        name = StringType(required=True)

    class Child(Parent):
        bio = StringType()

    input_data = {"bio": "Genius", "name": "Joey"}

    model = Child(input_data)
    model.validate()

    assert model.serialize() == input_data

    child = Child({"name": "Baby Jane", "bio": "Always behaves"})
    assert child.name == "Baby Jane"
    assert child.bio == "Always behaves"


def test_validation_uses_internal_state():
    class User(Model):
        name = StringType(required=True)
        age = IntType(required=True)

    u = User({"name": "Henry VIII"})
    u.age = 99
    u.validate()

    assert u.name == "Henry VIII"
    assert u.age == 99


def test_validation_fails_if_internal_state_is_invalid():
    class User(Model):
        status = StringType()
        name = StringType(required=True)
        age = IntType(required=True)

    u = User()
    with pytest.raises(DataError) as exception:
        u.validate()

    assert exception.value.errors, {
        "name": ["This field is required."],
        "age": ["This field is required."],
    }

    assert u.status is None
    with pytest.raises(UndefinedValueError):
        u.name == u.age


def test_returns_nice_conversion_errors():
    class User(Model):
        name = StringType(required=True)
        age = IntType(required=True)

    with pytest.raises(DataError) as exception:
        User({"name": "Jóhann", "age": "100 years"})

    errors = exception.value.errors

    assert errors == {
        "age": ["Value '100 years' is not int."],
    }


def test_returns_partial_data_with_conversion_errors():
    class User(Model):
        name = StringType(required=True)
        age = IntType(required=True)
        account_level = IntType()

    with pytest.raises(DataError) as exception:
        User({"name": "Jóhann", "age": "100 years", "account_level": "3"})

    partial_data = exception.value.partial_data

    assert partial_data == {
        "name": "Jóhann",
        "account_level": 3,
    }


def test_field_default():
    class User(Model):
        name = StringType(default="Doggy")

    u = User()
    assert User.name.__class__ == StringType
    assert u.name == "Doggy"


def test_attribute_default_to_none_if_no_value():
    class User(Model):
        name = StringType()

    u = User()
    assert u.name is None


def test_field_has_default_value():
    class Question(Model):
        question_id = StringType(required=True)

        type = StringType(default="text")

    q = Question(dict(question_id=1))

    assert q.type == "text"
    assert "type" in q
    assert q.get("type") == "text"


def test_default_value_when_updating_model():
    class Question(Model):
        question_id = StringType(required=True)

        type = StringType(default="text")

    q = Question(dict(question_id=1, type="not default"))
    assert q.type == "not default"

    q.validate(dict(question_id=2))
    assert q.type == "not default"


def test_explicit_values_override_defaults():
    class User(Model):
        name = StringType(default="Doggy")

    u = User({"name": "Voffi"})
    u.validate()
    assert u.name == "Voffi"

    u = User()
    u.name = "Guffi"
    u.validate()

    assert u.name == "Guffi"


def test_good_options_args():
    so = SchemaOptions(roles=None)
    assert so is not None

    assert so.roles == {}


def test_options_custom_args():
    class Foo(Model):
        class Options:
            _foo = "bar"

    f = Foo()
    assert f._schema.options._foo == "bar"


def test_options_custom_args_inheritance():
    class Foo(Model):
        class Options:
            _foo = "bar"

    class Moo(Foo):
        class Options:
            _bar = "baz"

    m = Moo()
    assert m._schema.options._foo == "bar"
    assert m._schema.options._bar == "baz"


def test_no_options_args():
    args = {}
    so = SchemaOptions(**args)
    assert so is not None


def test_options_parsing_from_model():
    class Foo(Model):
        class Options:
            namespace = "foo"
            roles = {}

    f = Foo()
    fo = f._schema.options

    assert fo.__class__ == SchemaOptions
    assert fo.namespace == "foo"
    assert fo.roles == {}


def test_options_parsing_from_optionsclass():
    class FooOptions(SchemaOptions):
        def __init__(self, **kwargs):
            kwargs["namespace"] = kwargs.get("namespace") or "foo"
            kwargs["roles"] = kwargs.get("roles") or {}
            super().__init__(**kwargs)

    class Foo(Model):
        __optionsclass__ = FooOptions

    f = Foo()
    fo = f._schema.options

    assert fo.__class__ == FooOptions
    assert fo.namespace == "foo"
    assert fo.roles == {}


def test_subclassing_preservers_roles():
    class Parent(Model):
        id = StringType()
        name = StringType()

        class Options:
            roles = {"public": blacklist("id")}

    class GrandParent(Parent):
        age = IntType()

    gramps = GrandParent({"id": "1", "name": "Edward", "age": 87})

    options = gramps._schema.options

    assert options.roles == {
        "public": blacklist("id"),
    }


def test_subclassing_overides_roles():
    class Parent(Model):
        id = StringType()
        gender = StringType()
        name = StringType()

        class Options:
            roles = {"public": blacklist("id", "gender"), "gender": blacklist("gender")}

    class GrandParent(Parent):
        age = IntType()
        family_secret = StringType()

        class Options:
            roles = {
                "grandchildren": whitelist("age"),
                "public": blacklist("id", "family_secret"),
            }

    gramps = GrandParent(
        {
            "id": "1",
            "name": "Edward",
            "gender": "Male",
            "age": 87,
            "family_secret": "Secretly Canadian",
        }
    )

    options = gramps._schema.options

    assert options.roles == {
        "grandchildren": whitelist("age"),
        "public": blacklist("id", "family_secret"),
        "gender": blacklist("gender"),
    }


def test_as_field_validate():
    class User(Model):
        name = StringType()

    class Card(Model):
        user = ModelType(User)

    c = Card({"user": {"name": "Doggy"}})
    assert c.user.name == "Doggy"

    with pytest.raises(ConversionError):
        c.user = [1]

    c.validate()
    assert (
        c.user.name == "Doggy"
    ), "Validation should not remove or modify existing data"


def test_model_field_validate_structure():
    class User(Model):
        name = StringType()

    class Card(Model):
        user = ModelType(User)

    with pytest.raises(DataError):
        Card({"user": [1, 2]})


def test_model_field_validate_only_when_field_is_set():
    class M0(Model):
        bar = StringType()

        def validate_bar(self, data, value):
            if data["bar"] and "bar" not in data["bar"]:
                raise ValidationError("Illegal value")

    class M1(Model):
        foo = StringType(required=True)

        def validate_foo(self, data, value):
            if "foo" not in data["foo"]:
                raise ValidationError("Illegal value")

    m = M0({})
    m.validate()

    m = M0({"bar": "foo"})
    with pytest.raises(DataError) as e:
        m.validate()
        assert isinstance(e["foo"][0], ErrorMessage)
        assert "Illegal value" in e["foo"][0]

    m = M0({"bar": "foobar"})
    m.validate()

    m = M1({})
    with pytest.raises(DataError) as e:
        m.validate()
        assert isinstance(e["foo"], ConversionError)
        assert "This field is required" in e["foo"]

    m = M1({"foo": "bar"})
    with pytest.raises(DataError) as e:
        m.validate()
        assert isinstance(e["foo"][0], ErrorMessage)
        assert "Illegal value" in e["foo"][0]

    m = M1({"foo": "foobar"})
    m.validate()


def test_model_deserialize_from_with_list():
    class User(Model):
        username = StringType(deserialize_from=["name", "user"])

    assert User({"name": "Ryan"}).username == "Ryan"
    assert User({"user": "Mike"}).username == "Mike"
    assert User({"username": "Mark"}).username == "Mark"
    assert (
        User({"username": "Mark", "name": "Second-class", "user": "key"}).username
        == "Mark"
    )


def test_model_deserialize_from_with_string():
    class User(Model):
        username = StringType(deserialize_from="name")

    assert User({"name": "Mike"}).username == "Mike"
    assert User({"username": "Mark"}).username == "Mark"
    assert User({"username": "Mark", "name": "Second-class field"}).username == "Mark"


def test_model_import_with_deserialize_mapping():
    class User(Model):
        username = StringType()

    mapping = {
        "username": ["name", "user"],
    }

    assert User({"name": "Ryan"}, deserialize_mapping=mapping).username == "Ryan"
    assert User({"user": "Mike"}, deserialize_mapping=mapping).username == "Mike"
    assert User({"username": "Mark"}, deserialize_mapping=mapping).username == "Mark"
    assert (
        User(
            {"username": "Mark", "name": "Second-class", "user": "key"},
            deserialize_mapping=mapping,
        ).username
        == "Mark"
    )


def test_model_import_data_with_mapping():
    class User(Model):
        username = StringType()

    mapping = {
        "username": ["name", "user"],
    }

    user = User()
    user.import_data({"name": "Ryan"}, mapping=mapping)
    assert user.username == "Ryan"


def test_nested_model_import_data_with_mappings():
    class Nested(Model):
        nested_attr = StringType()

    class Root(Model):
        root_attr = StringType()
        nxt_level = ModelType(Nested)

    mapping = {
        "root_attr": ["attr"],
        "nxt_level": ["next"],
        "model_mapping": {
            "nxt_level": {
                "nested_attr": ["attr"],
            },
        },
    }

    root = Root()
    root.import_data(
        {
            "attr": "root value",
            "next": {
                "attr": "nested value",
            },
        },
        mapping=mapping,
    )

    assert root.root_attr == "root value"
    assert root.nxt_level.nested_attr == "nested value"

    root = Root(
        {
            "attr": "root value",
            "next": {
                "attr": "nested value",
            },
        },
        deserialize_mapping=mapping,
    )

    assert root.root_attr == "root value"
    assert root.nxt_level.nested_attr == "nested value"


class SimpleModel(Model):
    field1 = StringType()
    field2 = StringType()


def test_keys():
    inst = SimpleModel({"field1": "foo", "field2": "bar"})

    assert inst.keys() == ["field1", "field2"]
    del inst.field2
    assert inst.keys() == ["field1"]


def test_values():
    inst = SimpleModel({"field1": "foo", "field2": "bar"})

    assert inst.values() == ["foo", "bar"]
    del inst.field2
    assert inst.values() == ["foo"]


def test_items():
    inst = SimpleModel({"field1": "foo", "field2": "bar"})

    assert inst.items() == [("field1", "foo"), ("field2", "bar")]
    del inst.field2
    assert inst.items() == [("field1", "foo")]


def test_iter():
    inst = SimpleModel({"field1": "foo", "field2": "bar"})

    assert [x for x in inst] == ["field1", "field2"]
    del inst.field2
    assert [x for x in inst] == ["field1"]


def test_membership():
    inst = SimpleModel({"field1": "foo", "field2": "bar"})

    assert "field1" in inst and "field2" in inst
    del inst.field2
    assert "field1" in inst and "field2" not in inst


def test_get():

    inst = SimpleModel({"field1": "foo"})
    assert inst.get("field1") == "foo"
    assert inst.get("foo") is None
    assert inst.get("foo", "bar") == "bar"

    inst = SimpleModel({"field1": "foo"}, init=False)
    assert inst.get("foo") is None


def test_getitem():

    inst = SimpleModel({"field1": "foo"})
    assert inst["field1"] == "foo"
    assert inst["field2"] is None
    with pytest.raises(KeyError):
        inst["foo"]

    inst = SimpleModel({"field1": "foo"}, init=False)
    assert inst["field1"] == "foo"
    with pytest.raises(UndefinedValueError):
        inst["field2"]
    with pytest.raises(UnknownFieldError):
        inst["foo"]


def test_setitem():
    inst = SimpleModel()

    with pytest.raises(KeyError):
        inst["foo"] = 1

    inst["field1"] = "foo"
    assert inst.field1 == "foo"


def test_delitem():
    inst = SimpleModel({"field1": "foo"})

    with pytest.raises(KeyError):
        del inst["foo"]

    del inst["field1"]
    with pytest.raises(AttributeError):
        inst.field1


def test_eq():
    inst = SimpleModel({"field1": "foo"})
    assert inst != "foo"


def test_repr():
    inst = SimpleModel({"field1": "foo"})
    assert repr(inst) == str(inst) == "<SimpleModel instance>"

    class FooModel(SimpleModel):
        def _repr_info(self):
            return str.join(", ", (self[k] for k in self))

    inst = FooModel({"field1": "foo", "field2": "bar"})
    assert repr(inst) == "<FooModel: foo, bar>"

    inst = FooModel({"field1": "é", "field2": "Ä"})
    assert repr(inst) == "<FooModel: é, Ä>"


def test_mock_recursive_model():
    class M(Model):
        m = ListType(ModelType("M", required=True), required=True)

    M.get_mock_object()


def test_append_field_to_model():
    class M(Model):
        a = IntType()

    M._append_field("b", StringType())

    input_data = {"a": 1, "b": "b"}

    m = M(input_data)
    assert m.b == "b"
    assert m.serialize() == input_data
